"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import math

from panda3d.core import LVecBase2i, PTAInt

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.gpu_command_queue import GPUCommandQueue
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.native import InternalLightManager, PointLight, ShadowManager
from metadrive.render_pipeline.rpcore.rpobject import RPObject

from metadrive.render_pipeline.rpcore.stages.apply_lights_stage import ApplyLightsStage
from metadrive.render_pipeline.rpcore.stages.collect_used_cells_stage import CollectUsedCellsStage
from metadrive.render_pipeline.rpcore.stages.cull_lights_stage import CullLightsStage
from metadrive.render_pipeline.rpcore.stages.flag_used_cells_stage import FlagUsedCellsStage
from metadrive.render_pipeline.rpcore.stages.shadow_stage import ShadowStage


class LightManager(RPObject):
    """ This class is a wrapper around the InternalLightManager, and provides
    additional functionality like setting up all required stages and defines."""

    # Maximum amount of lights, has to match the definitions in the native code
    MAX_LIGHTS = 65535

    # Maximum amount of shadow sources
    MAX_SOURCES = 2048

    def __init__(self, pipeline):
        """ Constructs the light manager """
        RPObject.__init__(self)
        self.pipeline = pipeline
        self.compute_tile_size()
        self.init_internal_manager()
        self.init_command_queue()
        self.init_shadow_manager()
        self.init_stages()

    @property
    def total_tiles(self):
        """ Returns the total amount of tiles """
        return self.num_tiles.x * self.num_tiles.y * \
            self.pipeline.settings["lighting.culling_grid_slices"]

    @property
    def num_lights(self):
        """ Returns the amount of stored lights """
        return self.internal_mgr.num_lights

    @property
    def num_shadow_sources(self):
        """ Returns the amount of stored shadow sources """
        return self.internal_mgr.num_shadow_sources

    @property
    def shadow_atlas_coverage(self):
        """ Returns the shadow atlas coverage in percentage  """
        return self.internal_mgr.shadow_manager.atlas.coverage * 100.0

    def add_light(self, light):
        """ Adds a new light """
        self.internal_mgr.add_light(light)
        self.pta_max_light_index[0] = self.internal_mgr.max_light_index

    def remove_light(self, light):
        """ Removes a light """
        self.internal_mgr.remove_light(light)
        self.pta_max_light_index[0] = self.internal_mgr.max_light_index

    def update(self):
        """ Main update method to process the GPU commands """
        self.internal_mgr.set_camera_pos(Globals.base.camera.get_pos(Globals.base.render))
        self.internal_mgr.update()
        self.shadow_manager.update()
        self.cmd_queue.process_queue()

    def reload_shaders(self):
        """ Reloads all assigned shaders """
        self.cmd_queue.reload_shaders()

    def compute_tile_size(self):
        """ Computes how many tiles there are on screen """
        self.tile_size = LVecBase2i(
            self.pipeline.settings["lighting.culling_grid_size_x"],
            self.pipeline.settings["lighting.culling_grid_size_y"]
        )
        num_tiles_x = int(math.ceil(Globals.resolution.x / float(self.tile_size.x)))
        num_tiles_y = int(math.ceil(Globals.resolution.y / float(self.tile_size.y)))
        self.debug(
            "Tile size =", self.tile_size.x, "x", self.tile_size.y, ", Num tiles =", num_tiles_x, "x", num_tiles_y
        )
        self.num_tiles = LVecBase2i(num_tiles_x, num_tiles_y)

    def init_command_queue(self):
        """ Inits the command queue """
        self.cmd_queue = GPUCommandQueue(self.pipeline)
        self.cmd_queue.register_input("LightData", self.img_light_data)
        self.cmd_queue.register_input("SourceData", self.img_source_data)
        self.internal_mgr.set_command_list(self.cmd_queue.command_list)

    def init_shadow_manager(self):
        """ Inits the shadow manager """
        self.shadow_manager = ShadowManager()
        self.shadow_manager.set_max_updates(self.pipeline.settings["shadows.max_updates"])
        self.shadow_manager.set_scene(Globals.base.render)
        self.shadow_manager.set_tag_state_manager(self.pipeline.tag_mgr)
        self.shadow_manager.atlas_size = self.pipeline.settings["shadows.atlas_size"]
        self.internal_mgr.shadow_manager = self.shadow_manager

    def init_shadows(self):
        """ Inits the shadows, this has to get called after the stages were
        created, because we need the GraphicsOutput of the shadow atlas, which
        is not available earlier """
        self.shadow_manager.set_atlas_graphics_output(self.shadow_stage.atlas_buffer)
        self.shadow_manager.init()

    def init_internal_manager(self):
        """ Creates the light storage manager and the buffer to store the light data """
        self.internal_mgr = InternalLightManager()
        self.internal_mgr.set_shadow_update_distance(self.pipeline.settings["shadows.max_update_distance"])

        # Storage for the Lights
        per_light_vec4s = 4
        self.img_light_data = Image.create_buffer("LightData", self.MAX_LIGHTS * per_light_vec4s, "RGBA16")
        self.img_light_data.clear_image()

        self.pta_max_light_index = PTAInt.empty_array(1)
        self.pta_max_light_index[0] = 0

        # Storage for the shadow sources
        per_source_vec4s = 5

        # IMPORTANT: RGBA32 is really required here. Otherwise artifacts and bad
        # shadow filtering occur due to precision issues
        self.img_source_data = Image.create_buffer("ShadowSourceData", self.MAX_SOURCES * per_source_vec4s, "RGBA32")
        self.img_light_data.clear_image()

        # Register the buffer
        inputs = self.pipeline.stage_mgr.inputs
        inputs["AllLightsData"] = self.img_light_data
        inputs["ShadowSourceData"] = self.img_source_data
        inputs["maxLightIndex"] = self.pta_max_light_index

    def init_stages(self):
        """ Inits all required stages for the lighting """

        add_stage = self.pipeline.stage_mgr.add_stage

        self.flag_cells_stage = FlagUsedCellsStage(self.pipeline)
        add_stage(self.flag_cells_stage)

        self.collect_cells_stage = CollectUsedCellsStage(self.pipeline)
        add_stage(self.collect_cells_stage)

        self.cull_lights_stage = CullLightsStage(self.pipeline)
        add_stage(self.cull_lights_stage)

        self.apply_lights_stage = ApplyLightsStage(self.pipeline)
        add_stage(self.apply_lights_stage)

        self.shadow_stage = ShadowStage(self.pipeline)
        self.shadow_stage.size = self.shadow_manager.get_atlas_size()
        add_stage(self.shadow_stage)

    def init_defines(self):
        """ Inits the common defines """
        defines = self.pipeline.stage_mgr.defines
        defines["LC_TILE_SIZE_X"] = self.tile_size.x
        defines["LC_TILE_SIZE_Y"] = self.tile_size.y
        defines["LC_TILE_SLICES"] = self.pipeline.settings["lighting.culling_grid_slices"]
        defines["LC_MAX_DISTANCE"] = self.pipeline.settings["lighting.culling_max_distance"]
        defines["LC_CULLING_SLICE_WIDTH"] = self.pipeline.settings["lighting.culling_slice_width"]
        defines["LC_MAX_LIGHTS_PER_CELL"] = self.pipeline.settings["lighting.max_lights_per_cell"]
        defines["SHADOW_ATLAS_SIZE"] = self.pipeline.settings["shadows.atlas_size"]

        # Register all light types as defines
        for attr in dir(PointLight):
            if attr.startswith("LT_"):
                defines[attr.upper()] = getattr(PointLight, attr)



""" Tool to distribute the rp and the panda3d build. """

from __future__ import print_function

import os
import sys
import shutil
from os.path import isdir, join, dirname, realpath, relpath

base_dir = realpath(dirname(__file__))
rp_dir = realpath(join(base_dir, "../../"))
os.chdir(base_dir)

sys.path.insert(0, rp_dir)
from rplibs.six.moves import input  # noqa # pylint: disable=import-error

# TODO: Add option to skip gui folders if debugger is disabled

rp_ignores = [

    # data
    "skybox-blend.zip",
    "skybox.jpg",
    "skybox-2.jpg",
    "default_cubemap/source",
    "default_cubemap/source_2",
    "default_cubemap/filter.compute.glsl",
    "default_cubemap/filter.py",
    "data/generate_txo_files.py",
    "README.md",
    "environment_brdf/generate_reference.py",
    "run_mitsuba.bat",
    ".mip",
    ".xml",
    ".exr",
    ".psd",
    ".diff",
    ".pyc",
    ".pdb",
    "__pycache__",
    "environment_brdf/res/",
    "film_grain/generate.py",
    "film_grain/grain.compute.glsl",
    "ies_profiles/PREVIEWS.jpg",
    "loading_screen_bg.png",

    # rpcore
    "native/scripts",
    "native/source",
    "native/win_amd",
    "native/win_i386",
    "native/.gitignore",
    "native/build.py",
    "native/CMakeLists",
    "native/update_module_builder.py",
    "native/config.ini",

    # rpplugins
    ".ffxml",
    "bloom/resources/SOURCE.txt",
    "bloom/resources/lens_dirt.png",
    "clouds/resources/generate_",
    "clouds/resources/noise.inc",
    "clouds/resources/precompute.py",
    "color_correction/resources/film_luts_raw",
    "color_correction/resources/generate_",
    "plugin_prefab",
    "scattering/resources/hosek_wilkie_scattering",
    "toolkit",
]

panda_ignores = [
    ".pdb",
    ".pyc",
    "python/.vs",
    "python/include",
]

app_ignores = [".pyc", ".blend"]


def copy_tree(source_dir, dest_dir, ignorelist, tree_pth):
    source = join(source_dir, tree_pth)
    dest = join(dest_dir, tree_pth)
    for basepath, dirnames, files in os.walk(source):
        for f in files:
            abspath = realpath(join(basepath, f))
            abspath = abspath.replace("\\", "/")

            for ignore in ignorelist:
                if ignore in abspath:
                    break
            else:
                local_pth = relpath(abspath, start=source)
                dest_pth = join(dest, local_pth)
                dname = dirname(dest_pth)
                if not isdir(dname):
                    print("Creating", dname)
                    os.makedirs(dname)
                shutil.copyfile(abspath, dest_pth)


def distribute():
    print("Render Pipeline Distributor v0.1")
    print("")
    print("Copying tree ..")

    dist_folder_name = "built"

    tmp_dir = join(base_dir, dist_folder_name, "render_pipeline")
    if isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir)

    for dname in ("config", "data", "rplibs", "effects", "rpcore", "rpplugins", "toolkit"):
        copy_tree(rp_dir, tmp_dir, rp_ignores, dname)

    shutil.copyfile(join(rp_dir, "LICENSE.txt"), join(tmp_dir, "LICENSE.txt"))

    print("Copying Panda3D build ...")

    python_pth = realpath(dirname(sys.executable))
    panda_pth = realpath(join(python_pth, ".."))

    tmp_dir = join(base_dir, dist_folder_name, "panda3d")
    if isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir)

    for dname in ("direct", "etc", "models", "panda3d", "pandac", "python"):
        copy_tree(panda_pth, tmp_dir, panda_ignores, dname)

    copy_tree(panda_pth, tmp_dir, panda_ignores + [".exe"], "bin")

    shutil.copyfile(join(panda_pth, "LICENSE"), join(tmp_dir, "LICENSE.txt"))

    # Copy launcher script
    shutil.copyfile(join(base_dir, "launch.templ.bat"), join(base_dir, dist_folder_name, "launch.bat"))

    # Copy application
    app_pth = join(base_dir, "../../../RenderPipeline-Samples/01-Material-Demo/")
    tmp_dir = join(base_dir, dist_folder_name, "application")

    copy_tree(app_pth, tmp_dir, app_ignores, ".")


if __name__ == "__main__":
    distribute()



import time
from collections import deque
from itertools import chain
from sys import getsizeof, stderr

from metadrive.engine.engine_utils import initialize_engine, get_engine, close_engine
from metadrive.envs import MetaDriveEnv

try:
    from reprlib import repr
except ImportError:
    pass


def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory foot# print an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers)  # user handlers take precedence
    seen = set()  # track which object id's have already been seen
    default_size = getsizeof(0)  # estimate sizeof object without __sizeof__

    def sizeof(o):

        if hasattr(o, "__dict__"):
            o = o.__dict__

        if id(o) in seen:  # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        # if verbose:
        # print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


# inner psutil function
def process_memory():
    import psutil
    import os
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss


def test_engine_memory_leak():
    try:

        default_config = MetaDriveEnv.default_config()
        default_config["map_config"]["config"] = 3

        close_engine()

        engine = initialize_engine(default_config)

        ct = time.time()
        last_lm = cm = process_memory()
        last_mem = 0.0
        for t in range(500):
            lt = time.time()

            engine.seed(0)

            engine = get_engine()

            nlt = time.time()
            lm = process_memory()
            # # print(
            #     "After {} Iters, Time {:.3f} Total Time {:.3f}, Memory Usage {:,} Memory Change {:,}".format(
            #         t + 1, nlt - lt, nlt - ct, lm - cm, lm - last_lm
            #     )
            # )
            last_lm = lm
            if t > 100:
                assert abs((lm - cm) - last_mem) < 10  # Memory should not have change > 1KB
            last_mem = lm - cm
    finally:
        close_engine()


def test_config_memory_leak():

    ct = time.time()
    last_lm = cm = process_memory()
    last_mem = 0.0
    for t in range(1000):
        lt = time.time()

        default_config = MetaDriveEnv.default_config()
        default_config.update({"map": 3, "type": "block_sequence", "config": 3})
        del default_config

        nlt = time.time()
        lm = process_memory()
        # print(
        #     "After {} Iters, Time {:.3f} Total Time {:.3f}, Memory Usage {:,} Memory Change {:,}".format(
        #         t + 1, nlt - lt, nlt - ct, lm - cm, lm - last_lm
        #     )
        # )
        last_lm = lm
        if t > 500:
            assert abs((lm - cm) - last_mem) < 10  # Memory should not have change > 1KB
        last_mem = lm - cm


if __name__ == "__main__":
    test_engine_memory_leak()
    test_config_memory_leak()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function

import time
import hashlib

from panda3d.core import PStatCollector, Mat4, Point4, Vec3
from metadrive.render_pipeline.rpcore.globals import Globals


def rgb_from_string(text, min_brightness=0.6):
    """ Creates a rgb color from a given string """
    ohash = hashlib.md5(text[::-1].encode("ascii")).hexdigest()
    r, g, b = int(ohash[0:2], 16), int(ohash[2:4], 16), int(ohash[4:6], 16)
    neg_inf = 1.0 - min_brightness
    return (
        min_brightness + r / 255.0 * neg_inf, min_brightness + g / 255.0 * neg_inf, min_brightness + b / 255.0 * neg_inf
    )


def profile(func):
    """ Handy decorator which can be used to profile a function with pstats """
    collector_name = "Debug:%s" % func.__name__

    global_showbase = Globals.base

    # Insert the collector to a custom dictionary attached to the base
    if hasattr(global_showbase, 'custom_collectors'):
        if collector_name in global_showbase.custom_collectors.keys():
            pstat = global_showbase.custom_collectors[collector_name]
        else:
            global_showbase.custom_collectors[collector_name] = \
                PStatCollector(collector_name)
            pstat = global_showbase.custom_collectors[collector_name]
    else:
        pstat = PStatCollector(collector_name)
        global_showbase.custom_collectors = {}
        global_showbase.custom_collectors[collector_name] = pstat

    def do_pstat(*args, **kargs):
        pstat.start()
        returned = func(*args, **kargs)
        pstat.stop()
        return returned

    do_pstat.__name__ = func.__name__
    do_pstat.__dict__ = func.__dict__
    do_pstat.__doc__ = func.__doc__
    return do_pstat


class profile_cpu(object):  # noqa # pylint: disable=invalid-name,too-few-public-methods
    """
    Context manager for profiling CPU duration. This is useful for timing
    loading of files or other CPU-heavy operations. Example usage:

      with profile_cpu("Some Task"):
        some_slow_operation()

    Duration of the process will be print out on the console later on.
    """
    def __init__(self, name):
        self.name = name

    def __enter__(self):
        self.start_time = time.process_time()

    def __exit__(self, *args):
        duration = (time.process_time() - self.start_time) * 1000.0
        print(self.name, "took", round(duration, 2), "ms ")


def snap_shadow_map(mvp, cam_node, resolution):
    """ 'Snaps' a shadow map to make sure it always is on full texel centers.
    This ensures no flickering occurs while moving the shadow map.
    This works by projecting the Point (0,0,0) to light space, compute the
    texcoord differences and offset the light world space position by that. """
    mvp = Mat4(mvp)
    base_point = mvp.xform(Point4(0, 0, 0, 1)) * 0.5 + 0.5
    texel_size = 1.0 / float(resolution)
    offset_x = base_point.x % texel_size
    offset_y = base_point.y % texel_size
    mvp.invert_in_place()
    new_base = mvp.xform(
        Point4(
            (base_point.x - offset_x) * 2.0 - 1.0, (base_point.y - offset_y) * 2.0 - 1.0, (base_point.z) * 2.0 - 1.0, 1
        )
    )
    cam_node.set_pos(cam_node.get_pos() - Vec3(new_base.x, new_base.y, new_base.z))



try:
    import logging
    import os
    import xml.etree.ElementTree as ET
    from pathlib import Path
    from typing import Any, Dict, List, Mapping, Tuple, Union, cast
    import argoverse
    from argoverse.data_loading.vector_map_loader import Node, append_additional_key_value_pair, \
        append_unique_key_value_pair, convert_node_id_list_to_xy, extract_node_waypt, get_lane_identifier, str_to_bool, \
        extract_node_from_ET_element
    from argoverse.map_representation.map_api import ArgoverseMap as AGMap
    from argoverse.map_representation.map_api import PITTSBURGH_ID, MIAMI_ID, ROOT
except ImportError:
    pass

from metadrive.component.argoverse_block.argoverse_block import ArgoverseBlock
from metadrive.component.lane.argoverse_lane import ArgoverseLane
from metadrive.component.map.base_map import BaseMap
from metadrive.constants import PGLineColor

logger = logging.getLogger(__name__)

_PathLike = Union[str, "os.PathLike[str]"]


class ArgoverseMap(BaseMap):
    """
    This class converting the Argoverse to MetaDrive and allow the interactive behaviors
    """

    # supported city mas
    SUPPORTED_MAPS = ["MIA", "PIT"]

    # block size
    BLOCK_LANE_NUM = 40

    def __init__(self, map_config, random_seed=0):
        # origin ag map
        self.AGMap = AGMap()
        map_config[self.SEED] = random_seed
        assert "city" in map_config, "City name is required when generating argoverse map"
        assert map_config["city"] in self.SUPPORTED_MAPS, "City generation of {} is not supported (We support {} now)". \
            format(map_config["city"], self.SUPPORTED_MAPS)
        self.city = map_config["city"]
        super(ArgoverseMap, self).__init__(map_config=map_config, random_seed=None)
        self.lane_id_lane = None

    def _generate(self):
        """
        Modified from argoverse-map api
        """
        city_name = self.config["city"]
        city_id = MIAMI_ID if city_name == "MIA" else PITTSBURGH_ID
        xml_fpath = Path(ROOT).resolve() / f"pruned_argoverse_{city_name}_{city_id}_vector_map.xml"
        tree = ET.parse(os.fspath(xml_fpath))
        root = tree.getroot()
        logger.info(f"Loaded root: {root.tag}")

        all_graph_nodes = {}
        lane_objs = {}
        # all children are either Nodes or Ways
        for child in root:
            if child.tag == "node":
                node_obj = extract_node_from_ET_element(child)
                all_graph_nodes[node_obj.id] = node_obj
            elif child.tag == "way":
                lane_obj, lane_id = self.extract_lane_segment_from_ET_element(child, all_graph_nodes)
                lane_objs[lane_id] = lane_obj
            else:
                logger.error("Unknown XML item encountered.")
                raise ValueError("Unknown XML item encountered.")
        lane_ids = self.AGMap.get_lane_ids_in_xy_bbox(
            *self.argoverse_position(self.config["center"]), self.config["city"], self.config["radius"]
        )
        self.lane_id_lane = lane_objs
        self._construct_road_network([lane_objs[k] for k in lane_ids])

    def _construct_road_network(self, lanes: list):
        # TODO split the blocks in the future, if we need the whole map
        chosen_lanes = lanes
        for lane in lanes:
            self._post_process_lane(lane)

        block = ArgoverseBlock(0, self.road_network, {lane.id: lane for lane in chosen_lanes})
        block.construct_block(self.engine.worldNP, self.engine.physics_world)
        self.blocks.append(block)

    def _post_process_lane(self, lane: ArgoverseLane):
        if lane.l_neighbor_id is not None and self.lane_id_lane[lane.l_neighbor_id].l_neighbor_id == lane.id:
            lane.line_color = (PGLineColor.YELLOW, PGLineColor.GREY)

    @staticmethod
    def extract_lane_segment_from_ET_element(child: ET.Element,
                                             all_graph_nodes: Mapping[int, Node]) -> Tuple[ArgoverseLane, int]:
        """
        Modified from Argoverse map-api
        """
        lane_dictionary: Dict[str, Any] = {}
        lane_id = get_lane_identifier(child)
        node_id_list: List[int] = []
        for element in child:
            # The cast on the next line is the result of a typeshed bug.  This really is a List and not a ItemsView.
            way_field = cast(List[Tuple[str, str]], list(element.items()))
            field_name = way_field[0][0]
            if field_name == "k":
                key = way_field[0][1]
                if key in {"predecessor", "successor"}:
                    append_additional_key_value_pair(lane_dictionary, way_field)
                else:
                    append_unique_key_value_pair(lane_dictionary, way_field)
            else:
                node_id_list.append(extract_node_waypt(way_field))

        lane_dictionary["centerline"] = convert_node_id_list_to_xy(node_id_list, all_graph_nodes)
        predecessors = lane_dictionary.get("predecessor", None)
        successors = lane_dictionary.get("successor", None)
        has_traffic_control = str_to_bool(lane_dictionary["has_traffic_control"])
        is_intersection = str_to_bool(lane_dictionary["is_intersection"])
        lnid = lane_dictionary["l_neighbor_id"]
        rnid = lane_dictionary["r_neighbor_id"]
        l_neighbor_id = None if lnid == "None" else int(lnid)
        r_neighbor_id = None if rnid == "None" else int(rnid)
        lane_segment = ArgoverseLane(
            str(node_id_list[0]),
            str(node_id_list[-1]),
            lane_id,
            has_traffic_control,
            lane_dictionary["turn_direction"],
            is_intersection,
            l_neighbor_id,
            r_neighbor_id,
            predecessors,
            successors,
            lane_dictionary["centerline"],
        )
        return lane_segment, lane_id

    @staticmethod
    def argoverse_position(pos):
        pos[1] *= -1
        return pos

    @staticmethod
    def metadrive_vector(pos):
        pos[1] *= -1
        return pos


if __name__ == "__main__":
    from metadrive.engine.engine_utils import initialize_engine
    from metadrive.envs.metadrive_env import MetaDriveEnv

    default_config = MetaDriveEnv.default_config()
    default_config["use_render"] = True
    default_config["debug"] = True
    engine = initialize_engine(default_config)

    # in argoverse coordinates
    xcenter, ycenter = 2599.5505965123866, 1200.0214763629717
    xmin = xcenter - 80  # 150
    xmax = xcenter + 80  # 150
    ymin = ycenter - 80  # 150
    ymax = ycenter + 80  # 150
    map = ArgoverseMap(
        {
            "city": "PIT",
            # "draw_map_resolution": 1024,
            "center": ArgoverseMap.metadrive_vector([xcenter, ycenter]),
            "radius": 100
        }
    )
    engine.map_manager.load_map(map)
    engine.enableMouse()

    # argoverse data set is as the same coordinates as panda3d
    engine.main_camera.set_bird_view_pos(ArgoverseMap.metadrive_vector([xcenter, ycenter]))
    while True:
        map.engine.step()



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

setup_logger(debug=True)

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "num_scenarios": 4,
            "traffic_density": 0.1,
            "start_seed": 3,
            "image_source": "mini_map",
            "manual_control": True,
            "use_render": True,
            "image_observation": False,
            "decision_repeat": 5,
            "rgb_clip": True,
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_NUM,
                BaseMap.GENERATE_CONFIG: 12,
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            }
        }
    )

    env.reset()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        env.render(text={"Frame": i, "Speed": env.vehicle.speed_km_h})
    env.close()



import pickle
from metadrive.utils.math import wrap_to_pi
import numpy as np

from metadrive import MultiAgentRoundaboutEnv
from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv
from metadrive.manager.traffic_manager import TrafficMode
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import setup_logger


def test_save_episode(vis=False):
    setup_logger(True)

    test_dump = True
    save_episode = True
    vis = vis
    env = SafeMetaDriveEnv(
        {
            "accident_prob": 0.8,
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "start_seed": 1000,
            # "manual_control": vis,
            "use_render": False,
            "agent_policy": IDMPolicy,
            "traffic_mode": TrafficMode.Trigger,
            "record_episode": save_episode,
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
                BaseMap.GENERATE_CONFIG: "CrCSC",
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            }
        }
    )
    step_info = []
    try:
        o, _ = env.reset()
        for i in range(1, 100000 if vis else 2000):
            step_info.append(
                {
                    name: [obj.position, obj.heading_theta, obj.class_name]
                    for name, obj in env.engine._spawned_objects.items()
                }
            )
            o, r, tm, tc, info = env.step([0, 1])
            if vis:
                env.render(mode="top_down", road_color=(35, 35, 35))
            if tm or tc:
                epi_info = env.engine.dump_episode("test_dump_single.pkl" if test_dump else None)
                break
        # with open("../test_export_record_scenario/test_dump_single.pkl", "rb") as f:
        env.config["replay_episode"] = epi_info
        env.config["record_episode"] = False
        o, _ = env.reset()
        for i in range(0, 100000 if vis else 2000):
            # if i % 5 ==0:
            for old_id, new_id in env.engine.replay_manager.record_name_to_current_name.items():
                obj = env.engine.replay_manager.spawned_objects[new_id]
                pos = obj.position
                heading = obj.heading_theta
                record_pos = env.engine.replay_manager.current_frame.step_info[old_id]["position"]
                record_heading = env.engine.replay_manager.current_frame.step_info[old_id]["heading_theta"]
                assert np.isclose(np.array([pos[0], pos[1], obj.get_z()]), np.array(record_pos)).all()
                assert abs(wrap_to_pi(heading - record_heading)) < 1e-2

                assert np.isclose(np.array([pos[0], pos[1]]), np.array(step_info[i][old_id][0]), 1e-2, 1e-2).all()
                assert abs(wrap_to_pi(heading - np.array(step_info[i][old_id][1]))) < 1e-2
            # assert abs(env.vehicle.get_z() - record_pos[-1]) < 1e-3
            o, r, tm, tc, info = env.step([0, 1])
            if vis:
                env.render()
            if info.get("replay_done", False):
                break
    finally:
        env.close()


def test_save_episode_marl(vis=False):
    """
    1. Set record_episode=True to record each episode
    2. dump_episode when terminated[__all__] == True
    3. You can keep recent episodes
    4. Input episode data to reset() function can replay the episode !
    """

    # setup_logger(True)

    test_dump = True
    dump_recent_episode = 5
    dump_count = 0
    env = MultiAgentRoundaboutEnv(
        dict(use_render=vis, manual_control=False, record_episode=True, horizon=100, force_seed_spawn_manager=True)
    )
    try:
        # Test Record
        o, _ = env.reset(seed=0)
        epi_info = None
        # for tt in range(10, 100):
        tt = 13
        # print("\nseed: {}\n".format(tt))
        env.engine.spawn_manager.seed(tt)
        o, _ = env.reset()
        for i in range(1, 100000 if vis else 600):
            o, r, tm, tc, info = env.step({agent_id: [0, .2] for agent_id in env.vehicles.keys()})
            if vis:
                env.render()
            if tm["__all__"]:
                epi_info = env.engine.dump_episode("test_dump.pkl")
                # test dump json
                # if test_dump:
                #     with open("test_dump_{}.json".format(dump_count), "w") as f:
                #         json.dump(epi_info, f)
                #     dump_count += 1
                #     dump_count = dump_count % dump_recent_episode
                break
                # env.reset()

        # with open("../test_export_record_scenario/test_dump.pkl", "rb") as epi_record:
        # input episode_info to restore
        env.config["replay_episode"] = epi_info

        env.config["record_episode"] = False
        o, _ = env.reset()
        for i in range(1, 100000 if vis else 2000):
            # if i % 5 ==0:
            for old_id, new_id in env.engine.replay_manager.record_name_to_current_name.items():
                obj = env.engine.replay_manager.spawned_objects[new_id]
                pos = obj.position
                heading = obj.heading_theta
                record_pos = env.engine.replay_manager.current_frame.step_info[old_id]["position"]
                record_heading = env.engine.replay_manager.current_frame.step_info[old_id]["heading_theta"]
                assert np.isclose(np.array([pos[0], pos[1], obj.get_z()]), np.array(record_pos)).all()
                assert abs(wrap_to_pi(heading - record_heading)) < 1e-2
            # print("Replay MARL step: {}".format(i))
            o, r, tm, tc, info = env.step({agent_id: [0, 0.1] for agent_id in env.vehicles.keys()})
            if vis:
                env.render()
            if tm["__all__"]:
                break
    finally:
        env.close()


if __name__ == "__main__":
    test_save_episode(vis=False)
    test_save_episode_marl(vis=False)



from panda3d.core import NodePath

from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import CreateRoadFrom, CreateAdverseRoad, ExtendStraightLane
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.constants import Decoration, PGLineType
from metadrive.engine.core.physics_world import PhysicsWorld
from metadrive.component.pg_space import ParameterSpace


class FirstPGBlock(PGBlock):
    """
    A special Set, only used to create the first block. One scene has only one first block!!!
    """
    NODE_1 = ">"
    NODE_2 = ">>"
    NODE_3 = ">>>"
    PARAMETER_SPACE = ParameterSpace({})
    ID = "I"
    SOCKET_NUM = 1
    ENTRANCE_LENGTH = 10

    def __init__(
        self,
        global_network: NodeRoadNetwork,
        lane_width: float,
        lane_num: int,
        render_root_np: NodePath,
        physics_world: PhysicsWorld,
        length: float = 30,
        ignore_intersection_checking=False
    ):
        place_holder = PGBlockSocket(Road(Decoration.start, Decoration.end), Road(Decoration.start, Decoration.end))
        super(FirstPGBlock, self).__init__(
            0, place_holder, global_network, random_seed=0, ignore_intersection_checking=ignore_intersection_checking
        )
        if length < self.ENTRANCE_LENGTH:
            print("Warning: first block length is two small", length, "<", self.ENTRANCE_LENGTH)
        self._block_objects = []
        basic_lane = StraightLane(
            [0, 0], [self.ENTRANCE_LENGTH, 0], line_types=(PGLineType.BROKEN, PGLineType.SIDE), width=lane_width
        )
        ego_v_spawn_road = Road(self.NODE_1, self.NODE_2)
        CreateRoadFrom(
            basic_lane,
            lane_num,
            ego_v_spawn_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        CreateAdverseRoad(
            ego_v_spawn_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        next_lane = ExtendStraightLane(basic_lane, length - self.ENTRANCE_LENGTH, [PGLineType.BROKEN, PGLineType.SIDE])
        other_v_spawn_road = Road(self.NODE_2, self.NODE_3)
        CreateRoadFrom(
            next_lane,
            lane_num,
            other_v_spawn_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        CreateAdverseRoad(
            other_v_spawn_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        self._create_in_world()

        # global_network += self.block_network
        global_network.add(self.block_network)

        socket = self.create_socket_from_positive_road(other_v_spawn_road)
        socket.set_index(self.name, 0)

        self.add_sockets(socket)
        self.attach_to_world(render_root_np, physics_world)
        self._respawn_roads = [other_v_spawn_road]

    def _try_plug_into_previous_block(self) -> bool:
        raise ValueError("BIG Recursive calculation error! Can not find a right sequence for building map! Check BIG")

    def destruct_block(self, physics_world: PhysicsWorld):
        """This block can not be destructed"""
        pass



from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import Config


class SafeMetaDriveEnv(MetaDriveEnv):
    def default_config(self) -> Config:
        config = super(SafeMetaDriveEnv, self).default_config()
        config.update(
            {
                "num_scenarios": 100,
                "accident_prob": 0.8,
                "traffic_density": 0.05,
                "crash_vehicle_done": False,
                "crash_object_done": False,
                "cost_to_reward": False,
            },
            allow_add_new_key=True
        )
        return config

    def __init__(self, config=None):
        super(SafeMetaDriveEnv, self).__init__(config)
        self.episode_cost = 0

    def reset(self, *args, **kwargs):
        self.episode_cost = 0
        return super(SafeMetaDriveEnv, self).reset(*args, **kwargs)

    def cost_function(self, vehicle_id: str):
        cost, step_info = super(SafeMetaDriveEnv, self).cost_function(vehicle_id)
        self.episode_cost += cost
        step_info["total_cost"] = self.episode_cost
        return cost, step_info


if __name__ == "__main__":
    env = SafeMetaDriveEnv(
        {
            "accident_prob": 0.5,
            "traffic_density": 0.15,
            # "manual_control": True,
            "agent_policy": IDMPolicy,
            "use_render": True,
            # "debug": True,
            'num_scenarios': 100,
            "start_seed": 129,
            # "traffic_density": 0.2,
            # "num_scenarios": 1,
            # # "start_seed": 187,
            # "out_of_road_cost": 1,
            # "debug": True,
            # "map": "X",
            # # "cull_scene": True,
            # "cost_to_reward":True,
            "vehicle_config": {
                "spawn_lane_index": (FirstPGBlock.NODE_2, FirstPGBlock.NODE_3, 2),
                # "show_lidar": True,
                # "show_side_detector": True,
                # "show_lane_line_detector": True,
                # "side_detector": dict(num_lasers=2, distance=50),  # laser num, distance
                # "lane_line_detector": dict(num_lasers=2, distance=20),  # laser num, distance
            }
        }
    )

    o, _ = env.reset()
    total_cost = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 0])
        total_cost += info["cost"]
        env.render(
            text={
                "cost": total_cost,
                "seed": env.current_seed,
                "reward": r,
                "total_cost": info["total_cost"],
            }
        )
        if info["crash_vehicle"]:
            print("crash_vehicle:cost {}, reward {}".format(info["cost"], r))
        if info["crash_object"]:
            print("crash_object:cost {}, reward {}".format(info["cost"], r))

        # if d:
        #     total_cost = 0
        #     print("done_cost:{}".format(info["cost"]), "done_reward;{}".format(r))
        #     print("Reset")
        #     env.reset()
    env.close()



from metadrive.examples import expert
from metadrive.policy.base_policy import BasePolicy


class ExpertPolicy(BasePolicy):
    def act(self, agent_id=None):
        action = expert(self.control_object)
        self.action_info["action"] = action
        return action



from nuplan.common.actor_state.tracked_objects_types import TrackedObjectType

from metadrive.component.traffic_participants.cyclist import Cyclist
from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.manager.base_manager import BaseManager
from metadrive.utils.nuplan.parse_object_state import parse_object_state


class NuplanParticipantManager(BaseManager):
    """
    This manager will control the walker and cyclist in the scenario
    """
    def __init__(self):
        raise DeprecationWarning("No all traffic participants are actuated by TrafficManager")
        super(NuplanParticipantManager, self).__init__()
        self.nuplan_id_to_obj = {}
        self._current_traffic_participants = None

    def after_reset(self):
        self._current_traffic_participants = self._get_current_traffic_participants()
        assert self.engine.episode_step == 0
        self.nuplan_id_to_obj = {}
        for nuplan_id, obj_state in self._current_traffic_participants[0].items():
            state = parse_object_state(obj_state, self.engine.current_map.nuplan_center)
            if obj_state.tracked_object_type == TrackedObjectType.BICYCLE or \
                    obj_state.tracked_object_type == TrackedObjectType.PEDESTRIAN:
                obj = self.spawn_object(
                    Cyclist if obj_state.tracked_object_type == TrackedObjectType.BICYCLE else Pedestrian,
                    position=state["position"],
                    heading_theta=state["heading"],
                )
                self.nuplan_id_to_obj[nuplan_id] = obj.name
                obj.set_velocity(state["velocity"])

    def after_step(self, *args, **kwargs):
        if self.episode_step >= self.current_scenario_length:
            return

        objs_to_eliminate = self.nuplan_id_to_obj.keys() - self._current_traffic_participants[self.engine.episode_step
                                                                                              ].keys()
        for nuplan_id in list(objs_to_eliminate):
            self.clear_objects([self.nuplan_id_to_obj[nuplan_id]])
            self.nuplan_id_to_obj.pop(nuplan_id)

        for nuplan_id, obj_state in self._current_traffic_participants[self.engine.episode_step].items():
            if obj_state.tracked_object_type != TrackedObjectType.PEDESTRIAN or \
                    obj_state.tracked_object_type != TrackedObjectType.BICYCLE:
                continue
            state = parse_object_state(obj_state, self.engine.current_map.nuplan_center)
            if nuplan_id in self.nuplan_id_to_obj and self.nuplan_id_to_obj[nuplan_id] in self.spawned_objects.keys():
                self.spawned_objects[self.nuplan_id_to_obj[nuplan_id]].set_position(state["position"])
                self.spawned_objects[self.nuplan_id_to_obj[nuplan_id]].set_heading_theta(state["heading"])
                self.spawned_objects[self.nuplan_id_to_obj[nuplan_id]].set_velocity(state["velocity"])
            else:
                obj = self.spawn_object(
                    Cyclist if obj_state.tracked_object_type == TrackedObjectType.BICYCLE else Pedestrian,
                    position=state["position"],
                    heading_theta=state["heading"],
                )
                self.nuplan_id_to_obj[nuplan_id] = obj.name
                obj.set_velocity(state["velocity"])

    @property
    def current_scenario(self):
        return self.engine.data_manager.current_scenario

    def _get_current_traffic_participants(self):
        length = self.engine.data_manager.current_scenario.get_number_of_iterations()
        detection_ret = {
            i: self.engine.data_manager.current_scenario.get_tracked_objects_at_iteration(i).tracked_objects
            for i in range(length)
        }
        for step, frame_data in detection_ret.items():
            new_frame_data = {}
            for obj in frame_data:
                new_frame_data[obj.track_token] = obj
            detection_ret[step] = new_frame_data
        return detection_ret

    @property
    def current_scenario_length(self):
        return self.engine.data_manager.current_scenario_length



# This test is broken for some reasons. Just remove it from CI temporarily.


def _test_remote_metadrive_env():
    from metadrive.envs.legacy_envs.remote_env import RemoteMetaDrive
    # Test
    envs = [RemoteMetaDrive(dict(map=7)) for _ in range(3)]
    ret = [env.reset() for env in envs]
    # print(ret)
    ret = [env.step(env.action_space.sample()) for env in envs]
    # print(ret)
    [env.reset() for env in envs]
    [env.close() for env in envs]
    # print('Success!')


if __name__ == '__main__':
    _test_remote_metadrive_env()



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from metadrive.third_party.kitsunetsuki.base.material import get_root_node, get_from_node


class TextureMixin(object):
    def get_diffuse(self, material, shader):
        for i in ('Color', 'Alpha'):
            # Image Texture [Color/Alpha] -> [Socket] Principled BSDF
            node = get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=shader, from_socket_name=i, to_socket_name='Base Color'
            )
            if node:
                return node

    def get_normal_map(self, material, shader):
        # Normal Map [Normal] -> [Normal] Principled BSDF
        normal_map = get_from_node(
            material.node_tree, 'NORMAL_MAP', to_node=shader, from_socket_name='Normal', to_socket_name='Normal'
        )
        if normal_map:
            # Image Texture [Color] -> [Color] Normal Map
            return get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=normal_map, from_socket_name='Color', to_socket_name='Color'
            )

    def get_emission_map(self, material, shader):
        # emission map pipeline
        # Image Texture [Color] -> [Emission Strength] Principled BSDF
        if shader.inputs['Emission Strength'].is_linked:
            node = get_from_node(
                material.node_tree,
                'TEX_IMAGE',
                to_node=shader,
                from_socket_name='Color',
                to_socket_name='Emission Strength'
            )
            if node:
                return node

        # emission color pipeline
        # Mix RGB [Color] -> [Emission] Principled BSDF
        mix_node = get_from_node(
            material.node_tree, 'MIX_RGB', to_node=shader, from_socket_name='Color', to_socket_name='Emission'
        )
        if mix_node:
            # Image Texture [Color] -> [Color1/Color2] Mix
            for input_ in mix_node.inputs:
                if input_.name.startswith('Color') and input_.is_linked:
                    return get_from_node(
                        material.node_tree,
                        'TEX_IMAGE',
                        to_node=mix_node,
                        from_socket_name='Color',
                        to_socket_name=input_.name
                    )
        else:
            # Image Texture [Color] -> [Emission] Principled BSDF
            return get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=shader, from_socket_name='Color', to_socket_name='Emission'
            )

    def get_specular_map(self, material, shader):
        # Math [Value] -> [Specular] Principled BSDF
        math_node = get_from_node(
            material.node_tree, 'MATH', to_node=shader, from_socket_name='Value', to_socket_name='Specular'
        )
        if math_node:
            # Image Texture [Color] -> [Input] Math
            return get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=math_node, from_socket_name='Color', to_socket_name='Value'
            )
        else:
            # Image Texture [Color] -> [Specular] Principled BSDF
            return get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=shader, from_socket_name='Color', to_socket_name='Specular'
            )

    def get_roughness_map(self, material, shader):
        # Math [Value] -> [Roughness] Principled BSDF
        math_node = get_from_node(
            material.node_tree, 'MATH', to_node=shader, from_socket_name='Value', to_socket_name='Roughness'
        )
        if math_node:
            # Image Texture [Color] -> [Input] Math
            return get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=math_node, from_socket_name='Color', to_socket_name='Value'
            )
        else:
            # Image Texture [Color] -> [Roughness] Principled BSDF
            return get_from_node(
                material.node_tree, 'TEX_IMAGE', to_node=shader, from_socket_name='Color', to_socket_name='Roughness'
            )

    def get_parallax_map(self, material, shader):
        return

    def make_texture(self, i, image_texture):
        raise NotImplementedError()

    def make_empty_texture(self, i):
        raise NotImplementedError()

    def get_images(self, material, shader):
        return tuple()

    def make_textures(self, material):
        results = []

        shader = None
        if material.node_tree is not None:
            output = get_root_node(material.node_tree, 'OUTPUT_MATERIAL')
            if output:
                shader = get_from_node(
                    material.node_tree,
                    'BSDF_PRINCIPLED',
                    to_node=output,
                    from_socket_name='BSDF',
                    to_socket_name='Surface'
                )

        if shader:
            image_textures = self.get_images(material, shader)
            last_texid = 0
            for i, (type_, image_texture) in enumerate(reversed(image_textures)):
                if image_texture is not None:
                    last_texid = len(image_textures) - i - 1
                    break

            for i, (type_, image_texture) in enumerate(image_textures):
                if image_texture is None:
                    if self._empty_textures:  # fill empty slot
                        result = self.make_empty_texture(type_)
                        results.append((type_, ) + result)
                    elif self._render_type == 'rp':
                        break

                elif image_texture:
                    result = self.make_texture(type_, image_texture)
                    results.append((type_, ) + result)

                if i >= last_texid and self._render_type == 'rp':
                    break

        return results



import numpy as np
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.envs.top_down_env import TopDownSingleFrameMetaDriveEnv, TopDownMetaDrive, TopDownMetaDriveEnvV2


def test_top_down_rendering():
    for env in [
            TopDownSingleFrameMetaDriveEnv(dict(num_scenarios=5, map="C", traffic_density=1.0)),
            TopDownMetaDrive(dict(num_scenarios=5, map="C", traffic_density=1.0)),
            TopDownMetaDrive(dict(num_scenarios=5, map="C", frame_stack=1, post_stack=2)),
            TopDownMetaDriveEnvV2(dict(num_scenarios=5, map="C", frame_stack=1, post_stack=2)),
            WaymoEnv(dict(num_scenarios=1, start_scenario_index=0)),
            WaymoEnv(dict(num_scenarios=1, start_scenario_index=1)),
            WaymoEnv(dict(num_scenarios=1, start_scenario_index=2)),
    ]:
        try:
            for _ in range(5):
                o, _ = env.reset()
                assert np.mean(o) > 0.0
                for _ in range(10):
                    o, *_ = env.step([0, 1])
                    assert np.mean(o) > 0.0
                for _ in range(10):
                    o, *_ = env.step([-0.05, 1])
                    assert np.mean(o) > 0.0
        finally:
            env.close()


def _vis_top_down_with_panda_render():
    env = TopDownMetaDrive(dict(use_render=True))
    try:
        o, _ = env.reset()
        for i in range(1000):
            o, r, tm, tc, i = env.step([0, 1])
            if tm or tc:
                break
    finally:
        env.close()


def _vis_top_down_with_panda_render_and_top_down_visualization():
    env = TopDownMetaDrive({"use_render": True})
    try:
        o, _ = env.reset()
        for i in range(2000):
            o, r, tm, tc, i = env.step([0, 1])
            if tm or tc:
                break
            env.render(mode="top_down")
    finally:
        env.close()


if __name__ == "__main__":
    # test_top_down_rendering()
    # _vis_top_down_with_panda_render()
    _vis_top_down_with_panda_render_and_top_down_visualization()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from panda3d.core import Vec3, CurveFitter


class SmoothConnectedCurve(object):
    """ Interface to a curve which also manages connecting the end of the
    curve with the beginning. """
    def __init__(self):
        self._curve = None
        self._modified = False
        self._border_points = 1
        self._color = (0, 0, 0)
        self._cv_points = [(0, 0), (0.5, 0), (1.0, 0)]
        self.build_curve()

    @property
    def modified(self):
        """ Returns whether the curve was modified since the creation """
        return self._modified

    @property
    def control_points(self):
        """ Returns a list of all controll points """
        return self._cv_points

    @control_points.setter
    def control_points(self, points):
        """ Sets the cv points to the given list of points """
        self._cv_points = points
        self._modified = True
        self.build_curve()

    @property
    def color(self):
        """ Returns the display color of the curve """
        return self._color

    @color.setter
    def color(self, rgb):
        """ Sets the display color of the curve """
        self._color = rgb

    def set_single_value(self, val):
        """ Sets the curve to be linear, and only use a single value """
        self._cv_points = [[0.5, val]]
        self._modified = False
        self.build_curve()

    def append_cv(self, x, y):
        """ Appends a new cv and returns the index of the attached cv """
        self._cv_points.append([x, y])
        self.build_curve()
        self._modified = True
        return len(self._cv_points) - 1

    def remove_cv(self, index):
        """ Attempts to remove the cv at the given index, does nothing if only
        one control point is left """
        if len(self._cv_points) > 1:
            del self._cv_points[index]
        self._modified = True
        self.build_curve()

    def build_curve(self):
        """ Rebuilds the curve based on the controll point values """
        sorted_points = sorted(self._cv_points, key=lambda v: v[0])
        first_point = sorted_points[0]
        fitter = CurveFitter()

        # Duplicate curve at the beginning
        for i in range(self._border_points):
            end_point = self._cv_points[(-i + self._border_points - 1) % len(self._cv_points)]
            end_point = first_point
            fitter.add_xyz(0.0, Vec3(0, end_point[1], 0))

        # Append the actual points
        for point in self._cv_points:
            # Clamp point x position to avoid artifacts at the beginning
            point_t = max(0.01, point[0])
            fitter.add_xyz(point_t, Vec3(point_t, point[1], 0))

        # Duplicate curve at the end
        for i in range(self._border_points):
            start_point = self._cv_points[i % len(self._cv_points)]
            start_point = first_point
            fitter.add_xyz(1.0, Vec3(1, start_point[1], 0))

        fitter.sort_points()
        fitter.wrap_hpr()
        fitter.compute_tangents(1.0)

        self._curve = fitter.make_hermite()

    def set_cv_value(self, index, x_value, y_value):
        """ Updates the cv point at the given index """
        self._cv_points[index] = [x_value, y_value]
        self._modified = True

    def get_value(self, offset):
        """ Returns the value on the curve ranging whereas the offset should be
        from 0 to 1 (0 denotes the start of the curve). The returned value will
        be a value from 0 to 1 as well. """
        point = Vec3(0)
        self._curve.evaluate_xyz(offset, point)
        return max(0.0, min(1.0, point.y))

    def serialize(self):
        """ Returns the value of the curve as yaml list """
        points = ["[{:5.10f},{:5.10f}]".format(float(a), float(b)) for a, b in self._cv_points]
        return "[" + ','.join(points) + "]"



import math
import numpy as np
from direct.controls.InputState import InputState

from metadrive.utils import is_win, is_mac

if (not is_win()) and (not is_mac()):
    try:
        import evdev
        from evdev import ecodes, InputDevice
    except ImportError:
        pass

from metadrive.utils import import_pygame

pygame = import_pygame()


class Controller:
    def process_input(self, vehicle):
        raise NotImplementedError

    def process_others(self, *args, **kwargs):
        pass


class KeyboardController(Controller):
    STEERING_INCREMENT = 0.04
    STEERING_DECAY = 0.25

    THROTTLE_INCREMENT = 0.1
    THROTTLE_DECAY = 0.2

    BRAKE_INCREMENT = 0.5
    BRAKE_DECAY = 0.5

    def __init__(self, pygame_control):
        self.pygame_control = pygame_control
        if self.pygame_control:
            pygame.init()
        else:
            self.inputs = InputState()
            self.inputs.watchWithModifiers('forward', 'w')
            self.inputs.watchWithModifiers('reverse', 's')
            self.inputs.watchWithModifiers('turnLeft', 'a')
            self.inputs.watchWithModifiers('turnRight', 'd')
        self.steering = 0.
        self.throttle_brake = 0.
        self.np_random = np.random.RandomState(None)

    def process_input(self, vehicle):
        if not self.pygame_control:
            steering = 0.
            throttle_brake = 0.
            if not self.inputs.isSet('turnLeft') and not self.inputs.isSet('turnRight'):
                steering = 0.
            else:
                if self.inputs.isSet('turnLeft'):
                    steering = 1.0
                if self.inputs.isSet('turnRight'):
                    steering = -1.0
            if not self.inputs.isSet('forward') and not self.inputs.isSet("reverse"):
                throttle_brake = 0.
            else:
                if self.inputs.isSet('forward'):
                    throttle_brake = 1.0
                if self.inputs.isSet('reverse'):
                    throttle_brake = -1.0
        else:
            steering = 0.
            throttle_brake = 0.
            key_press = pygame.key.get_pressed()
            throttle_brake += key_press[pygame.K_w] - key_press[pygame.K_s]
            steering += key_press[pygame.K_a] - key_press[pygame.K_d]

        self.further_process(steering, throttle_brake)

        return np.array([self.steering, self.throttle_brake], dtype=np.float64)

    def further_process(self, steering, throttle_brake):
        if steering == 0.:
            if self.steering > 0.:
                self.steering -= self.STEERING_DECAY
                self.steering = max(0., self.steering)
            elif self.steering < 0.:
                self.steering += self.STEERING_DECAY
                self.steering = min(0., self.steering)
        if throttle_brake == 0.:
            if self.throttle_brake > 0.:
                self.throttle_brake -= self.THROTTLE_DECAY
                self.throttle_brake = max(self.throttle_brake, 0.)
            elif self.throttle_brake < 0.:
                self.throttle_brake += self.BRAKE_DECAY
                self.throttle_brake = min(0., self.throttle_brake)

        if steering > 0.:
            self.steering += self.STEERING_INCREMENT if self.steering > 0. else self.STEERING_DECAY
        elif steering < 0.:
            self.steering -= self.STEERING_INCREMENT if self.steering < 0. else self.STEERING_DECAY

        if throttle_brake > 0.:
            self.throttle_brake = max(self.throttle_brake, 0.)
            self.throttle_brake += self.THROTTLE_INCREMENT
        elif throttle_brake < 0.:
            self.throttle_brake = min(self.throttle_brake, 0.)
            self.throttle_brake -= self.BRAKE_INCREMENT

        rand = self.np_random.rand() / 10000
        # self.throttle_brake += rand[0]
        self.steering += rand

        self.throttle_brake = min(max(-1., self.throttle_brake), 1.)
        self.steering = min(max(-1., self.steering), 1.)

    def process_others(self, takeover_callback=None):
        """This function allows the outer loop to call callback if some signal is received by the controller."""
        if (takeover_callback is None) or (not self.pygame_control) or (not pygame.get_init()):
            return
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN and event.key == pygame.K_t:
                # Here we allow user to press T for takeover callback.
                takeover_callback()


class SteeringWheelController(Controller):
    RIGHT_SHIFT_PADDLE = 4
    LEFT_SHIFT_PADDLE = 5
    STEERING_MAKEUP = 1.5

    def __init__(self):
        try:
            import evdev
            from evdev import ecodes, InputDevice
        except ImportError:
            print(
                "Fail to load evdev, which is required for steering wheel control. "
                "Install evdev via pip install evdev"
            )
        pygame.display.init()
        pygame.joystick.init()
        assert not is_win(), "Joystick is supported in linux and mac only"
        assert pygame.joystick.get_count() > 0, "Please connect joystick or use keyboard input"
        print("Successfully Connect your Joystick!")

        ffb_device = evdev.list_devices()[0]
        self.ffb_dev = InputDevice(ffb_device)

        self.joystick = pygame.joystick.Joystick(0)
        self.joystick.init()

        self.right_shift_paddle = False
        self.left_shift_paddle = False

        self.button_circle = False
        self.button_rectangle = False
        self.button_triangle = False
        self.button_x = False

        self.button_up = False
        self.button_down = False
        self.button_right = False
        self.button_left = False

    def process_input(self, vehicle):
        pygame.event.pump()
        steering = -self.joystick.get_axis(0)
        throttle_brake = -self.joystick.get_axis(2) + self.joystick.get_axis(3)
        offset = 30
        val = int(65535 * (vehicle.speed_km_h + offset) / (120 + offset)) if vehicle is not None else 0
        self.ffb_dev.write(ecodes.EV_FF, ecodes.FF_AUTOCENTER, val)
        self.right_shift_paddle = True if self.joystick.get_button(self.RIGHT_SHIFT_PADDLE) else False
        self.left_shift_paddle = True if self.joystick.get_button(self.LEFT_SHIFT_PADDLE) else False

        self.left_shift_paddle = True if self.joystick.get_button(self.LEFT_SHIFT_PADDLE) else False
        self.left_shift_paddle = True if self.joystick.get_button(self.LEFT_SHIFT_PADDLE) else False

        self.button_circle = True if self.joystick.get_button(2) else False
        self.button_rectangle = True if self.joystick.get_button(1) else False
        self.button_triangle = True if self.joystick.get_button(3) else False
        self.button_x = True if self.joystick.get_button(0) else False

        hat = self.joystick.get_hat(0)
        self.button_up = True if hat[-1] == 1 else False
        self.button_down = True if hat[-1] == -1 else False
        self.button_left = True if hat[0] == -1 else False
        self.button_right = True if hat[0] == 1 else False

        return [steering * self.STEERING_MAKEUP, throttle_brake / 2]


class XboxController(Controller):
    """Control class for Xbox wireless controller
    Accept both wired and wireless connection
    Max steering, throttle, and break are bound by _discount
    """
    STEERING_DISCOUNT = 0.5
    THROTTLE_DISCOUNT = 0.5
    BREAK_DISCOUNT = 0.5
    BUTTON_X_MAP = 2
    BUTTON_Y_MAP = 3
    BUTTON_A_MAP = 0
    BUTTON_B_MAP = 1
    TRIGGER_RIGHT_MAP = 5
    TRIGGER_LEFT_MAP = 2

    def __init__(self):
        try:
            import evdev
            from evdev import ecodes, InputDevice
        except ImportError:
            print(
                "Fail to load evdev, which is required for steering wheel control. "
                "Install evdev via pip install evdev"
            )
        pygame.display.init()
        pygame.joystick.init()
        assert not is_win(), "Joystick is supported in linux and mac only"
        assert pygame.joystick.get_count() > 0, "Please connect joystick or use keyboard input"
        print("Successfully Connect your Joystick!")

        self.joystick = pygame.joystick.Joystick(0)
        self.joystick.init()

        self.button_x = False
        self.button_y = False
        self.button_a = False
        self.button_b = False

        self.button_up = False
        self.button_down = False
        self.button_right = False
        self.button_left = False

    def process_input(self, vehicle):
        pygame.event.pump()
        steering = -self.joystick.get_axis(0)
        if abs(steering) < 0.05:
            steering = 0
        elif steering < 0:
            steering = -(math.pow(2, abs(steering) * self.STEERING_DISCOUNT) - 1)
        else:
            steering = math.pow(2, abs(steering) * self.STEERING_DISCOUNT) - 1
        raw_throttle = self.joystick.get_axis(self.TRIGGER_RIGHT_MAP)
        raw_brake = self.joystick.get_axis(self.TRIGGER_LEFT_MAP)
        # 1+raw_throttle will map throttle between 0,2 need *0.5 to bound it between 0,1
        throttle = (1 + raw_throttle) * 0.5 * self.THROTTLE_DISCOUNT
        brake = (1 + raw_brake) * 0.5 * self.BREAK_DISCOUNT
        throttle_brake = throttle - brake

        self.button_x = True if self.joystick.get_button(self.BUTTON_X_MAP) else False
        self.button_y = True if self.joystick.get_button(self.BUTTON_Y_MAP) else False
        self.button_a = True if self.joystick.get_button(self.BUTTON_A_MAP) else False
        self.button_b = True if self.joystick.get_button(self.BUTTON_B_MAP) else False

        hat = self.joystick.get_hat(0)
        self.button_up = True if hat[-1] == 1 else False
        self.button_down = True if hat[-1] == -1 else False
        self.button_left = True if hat[0] == -1 else False
        self.button_right = True if hat[0] == 1 else False

        return [steering, throttle_brake]



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import struct
from panda3d.core import TransformState, Vec3, Mat4, BoundingSphere

from metadrive.render_pipeline.rpcore.rpobject import RPObject


class EnvironmentProbe(RPObject):
    """ Simple class, representing an environment probe """
    def __init__(self):
        """ Inits a new environment probe """
        RPObject.__init__(self)
        self.index = -1
        self.last_update = -1
        self._transform = TransformState.make_identity()
        self._bounds = BoundingSphere(Vec3(0), 1.0)
        self._modified = True
        self._parallax_correction = True
        self._border_smoothness = 0.1

    @property
    def modified(self):
        """ Returns whether the probe was modified since the last write """
        return self._modified

    @property
    def bounds(self):
        """ Returns the probes worldspace bounds """
        return self._bounds

    @property
    def parallax_correction(self):
        """ Returns whether parallax correction is enabled for this probe """
        return self._parallax_correction

    @parallax_correction.setter
    def parallax_correction(self, value):
        """ Sets whether parallax correction is enabled for this probe """
        self._parallax_correction = value
        self._modified = True

    @property
    def border_smoothness(self):
        """ Returns the border smoothness factor """
        return self._border_smoothness

    @border_smoothness.setter
    def border_smoothness(self, value):
        """ Sets the border smoothness factor """
        self._border_smoothness = value
        self._modified = True

    def set_pos(self, *args):
        """ Sets the probe position """
        self._transform = self._transform.set_pos(Vec3(*args))
        self.update_bounds()

    def set_hpr(self, *args):
        """ Sets the probe rotation """
        self._transform = self._transform.set_hpr(Vec3(*args))
        self.update_bounds()

    def set_scale(self, *args):
        """ Sets the probe scale """
        self._transform = self._transform.set_scale(Vec3(*args))
        self.update_bounds()

    def set_mat(self, matrix):
        """ Sets the probes matrix, overrides all other transforms """
        self._transform = TransformState.make_mat(matrix)
        self.update_bounds()

    def update_bounds(self):
        """ Updates the spheres bounds """
        mat = self._transform.get_mat()
        mid_point = mat.xform_point(Vec3(0, 0, 0))
        max_point = mat.xform_point(Vec3(1, 1, 1))
        radius = (mid_point - max_point).length()
        self._bounds = BoundingSphere(mid_point, radius)
        self._modified = True

    @property
    def matrix(self):
        """ Returns the matrix of the probe """
        return self._transform.get_mat()

    def write_to_buffer(self, buffer_ptr):
        """ Writes the probe to a given byte buffer """
        data, mat = [], Mat4(self._transform.get_mat())
        mat.invert_in_place()
        for i in range(4):
            for j in range(3):
                data.append(mat.get_cell(i, j))
        data += (float(self.index), 1.0 if self._parallax_correction else 0.0, self._border_smoothness, 0)
        data.append(self._bounds.get_center().x)
        data.append(self._bounds.get_center().y)
        data.append(self._bounds.get_center().z)
        data.append(self._bounds.get_radius())
        byte_data = struct.pack("f" * 20, *data)

        # 4 = sizeof float, 20 = floats per cubemap
        bytes_per_probe = 4 * 20
        buffer_ptr.set_subdata(self.index * bytes_per_probe, bytes_per_probe, byte_data)

        self._modified = False



from metadrive.base_class.randomizable import Randomizable


class BaseManager(Randomizable):
    """
    Managers should be created and registered after launching BaseEngine
    TODO LQY, inherit event manager so that we can control the process in a standard way
    """
    PRIORITY = 10  # the engine will call managers according to the priority

    def __init__(self):
        from metadrive.engine.engine_utils import get_engine, engine_initialized
        assert engine_initialized(), "You should not create manager before the initialization of BaseEngine"
        # self.engine = get_engine()
        Randomizable.__init__(self, get_engine().global_random_seed)
        self.spawned_objects = {}

    @property
    def episode_step(self):
        return self.engine.episode_step

    def before_step(self, *args, **kwargs) -> dict:
        """
        Usually used to set actions for all elements with their policies
        """
        return dict()

    def step(self, *args, **kwargs):
        pass

    def after_step(self, *args, **kwargs) -> dict:
        """
        Update state for this manager after system advancing dt
        """
        return dict()

    def before_reset(self):
        """
        Update episode level config to this manager and clean element or detach element
        """
        self.clear_objects([object_id for object_id in self.spawned_objects.keys()])
        self.spawned_objects = {}

    def reset(self):
        """
        Generate objects according to some pre-defined rules
        """
        pass

    def after_reset(self):
        """
        Usually used to record information after all managers called reset(),
        Since reset() of managers may influence each other
        """
        pass

    def destroy(self):
        """
        Destroy manager
        """
        # self.engine = None
        self.spawned_objects = None

    def spawn_object(self, object_class, **kwargs):
        """
        Spawn one objects
        """
        object = self.engine.spawn_object(object_class, **kwargs)
        self.spawned_objects[object.id] = object
        return object

    def clear_objects(self, *args, **kwargs):
        """
        Same as the function in engine, clear objects, Return exclude object ids

        filter: A list of object ids or a function returning a list of object id
        """
        exclude_objects = self.engine.clear_objects(*args, **kwargs)
        for obj in exclude_objects:
            self.spawned_objects.pop(obj)
        return exclude_objects

    def get_objects(self, *args, **kwargs):
        return self.engine.get_objects(*args, *kwargs)

    def change_object_name(self, obj, new_name):
        """
        Change the name of one object, Note: it may bring some bugs if abusing!
        """
        self.engine.change_object_name(obj, new_name)
        obj = self.spawned_objects.pop(obj.name)
        self.spawned_objects[new_name] = obj
        obj.name = new_name

    def add_policy(self, object_id, policy_class, *policy_args, **policy_kwargs):
        return self.engine.add_policy(object_id, policy_class, *policy_args, **policy_kwargs)

    def get_policy(self, object_id):
        return self.engine.get_policy(object_id)

    def has_policy(self, object_id, policy_cls=None):
        return self.engine.has_policy(object_id, policy_cls)

    def get_state(self):
        """This function will be called by RecordManager to collect manager state, usually some mappings"""
        return {"spawned_objects": {name: v.class_name for name, v in self.spawned_objects.items()}}

    def set_state(self, state: dict, old_name_to_current=None):
        """
        A basic function for restoring spawned objects mapping
        """
        assert self.episode_step == 0, "This func can only be called after env.reset() without any env.step() called"
        if old_name_to_current is None:
            old_name_to_current = {key: key for key in state.keys()}
        spawned_objects = state["spawned_objects"]
        ret = {}
        for name, class_name in spawned_objects.items():
            current_name = old_name_to_current[name]
            name_obj = self.engine.get_objects([current_name])
            assert current_name in name_obj and name_obj[current_name
                                                         ].class_name == class_name, "Can not restore mappings!"
            ret[current_name] = name_obj[current_name]
        self.spawned_objects = ret

    @property
    def class_name(self):
        return self.__class__.__name__

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()

    def get_metadata(self):
        """
        This function will store the metadata of each manager before the episode start, usually, we put some raw real
        world data in it, so that we won't lose information
        """
        assert self.episode_step == 0, "This func can only be called after env.reset() without any env.step() called"
        return {}



"""
We implement this top-down view render based on eleurent/Highway-Env
See more information on its Github page: https://github.com/eleurent/highway-env
"""
import os
import sys
from typing import Tuple

import gymnasium as gym
import numpy as np

from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.constants import Decoration, DEFAULT_AGENT, EDITION
from metadrive.obs.observation_base import ObservationBase
from metadrive.obs.top_down_obs_impl import WorldSurface, ObservationWindow, COLOR_BLACK, \
    VehicleGraphics, LaneGraphics
from metadrive.utils import import_pygame

pygame = import_pygame()


class TopDownObservation(ObservationBase):
    """
    Most of the source code is from Highway-Env, we only optimize and integrate it in MetaDrive
    See more information on its Github page: https://github.com/eleurent/highway-env
    """
    RESOLUTION = (200, 200)  # pix x pix
    MAP_RESOLUTION = (2000, 2000)  # pix x pix

    # MAX_RANGE = (50, 50)  # maximum detection distance = 50 M

    def __init__(self, vehicle_config, clip_rgb: bool, onscreen, resolution=None, max_distance=50):
        self.resolution = resolution or self.RESOLUTION
        super(TopDownObservation, self).__init__(vehicle_config)
        self.rgb_clip = clip_rgb
        self.num_stacks = 3

        # self.obs_shape = (64, 64)
        self.obs_shape = self.resolution

        self.pygame = import_pygame()

        self.onscreen = onscreen
        main_window_position = (0, 0)

        self._center_pos = None  # automatically change, don't set the value
        self._should_draw_map = True
        self._canvas_to_display_scaling = 0.0
        self.max_distance = max_distance

        # scene
        self.road_network = None
        # self.engine = None

        # initialize
        pygame.init()
        pygame.display.set_caption(EDITION + " (Top-down)")
        # main_window_position means the left upper location.
        os.environ['SDL_VIDEO_WINDOW_POS'] = '{},{}' \
            .format(main_window_position[0] - self.resolution[0], main_window_position[1])
        # Used for display only!
        self.screen = pygame.display.set_mode(
            (self.resolution[0] * 2, self.resolution[1] * 2)
        ) if self.onscreen else None

        # canvas
        self.init_canvas()
        self.init_obs_window()

    def init_obs_window(self):
        self.obs_window = ObservationWindow((self.max_distance, self.max_distance), self.resolution)

    def init_canvas(self):
        self.canvas_runtime = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))
        self.canvas_background = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))

    def reset(self, env, vehicle=None):
        # self.engine = env.engine
        self.road_network = env.current_map.road_network
        self._should_draw_map = True

    def render(self) -> np.ndarray:
        if self.onscreen:
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()

        if self._should_draw_map:
            self.draw_map()

        self.draw_scene()

        if self.onscreen:
            self.screen.fill(COLOR_BLACK)
            screen = self.obs_window.get_screen_window()
            if screen.get_size() == self.screen.get_size():
                self.screen.blit(screen, (0, 0))
            else:
                pygame.transform.smoothscale(self.obs_window.get_screen_window(), self.screen.get_size(), self.screen)
            pygame.display.flip()

    def get_screenshot(self, name="screenshot.png"):
        pygame.image.save(self.screen, name)

    def draw_map(self) -> pygame.Surface:
        """
        :return: a big map surface, clip  and rotate to use a piece of it
        """

        # Setup the maximize size of the canvas
        # scaling and center can be easily found by bounding box

        # TODO(pzh) We can reuse function draw_top_down_map here!

        b_box = self.road_network.get_bounding_box()
        self.canvas_background.fill(COLOR_BLACK)
        self.canvas_runtime.fill(COLOR_BLACK)
        self.canvas_background.set_colorkey(self.canvas_background.BLACK)
        x_len = b_box[1] - b_box[0]
        y_len = b_box[3] - b_box[2]
        max_len = max(x_len, y_len) + 20  # Add more 20 meters
        scaling = self.MAP_RESOLUTION[1] / max_len - 0.1
        assert scaling > 0

        # real-world distance * scaling = pixel in canvas
        self.canvas_background.scaling = scaling
        self.canvas_runtime.scaling = scaling
        # self._scaling = scaling

        centering_pos = ((b_box[0] + b_box[1]) / 2, (b_box[2] + b_box[3]) / 2)
        # self._center_pos = centering_pos
        self.canvas_runtime.move_display_window_to(centering_pos)

        self.canvas_background.move_display_window_to(centering_pos)
        for _from in self.road_network.graph.keys():
            decoration = True if _from == Decoration.start else False
            for _to in self.road_network.graph[_from].keys():
                for l in self.road_network.graph[_from][_to]:
                    two_side = True if l is self.road_network.graph[_from][_to][-1] or decoration else False
                    LaneGraphics.LANE_LINE_WIDTH = 0.5
                    LaneGraphics.display(l, self.canvas_background, two_side)

        self.obs_window.reset(self.canvas_runtime)

        self._should_draw_map = False

    def draw_scene(self):
        # Set the active area that can be modify to accelerate
        assert len(self.engine.agents) == 1, "Don't support multi-agent top-down observation yet!"
        vehicle = self.engine.agents[DEFAULT_AGENT]
        pos = self.canvas_runtime.pos2pix(*vehicle.position)
        clip_size = (int(self.obs_window.get_size()[0] * 1.1), int(self.obs_window.get_size()[0] * 1.1))
        self.canvas_runtime.set_clip((pos[0] - clip_size[0] / 2, pos[1] - clip_size[1] / 2, clip_size[0], clip_size[1]))
        self.canvas_runtime.fill(COLOR_BLACK)
        self.canvas_runtime.blit(self.canvas_background, (0, 0))

        # Draw vehicles
        # TODO PZH: I hate computing these in pygame-related code!!!
        ego_heading = vehicle.heading_theta
        ego_heading = ego_heading if abs(ego_heading) > 2 * np.pi / 180 else 0

        VehicleGraphics.display(
            vehicle=vehicle, surface=self.canvas_runtime, heading=ego_heading, color=VehicleGraphics.GREEN
        )
        for v in self.engine.traffic_manager.vehicles:
            if v is vehicle:
                continue
            h = v.heading_theta
            h = h if abs(h) > 2 * np.pi / 180 else 0
            VehicleGraphics.display(vehicle=v, surface=self.canvas_runtime, heading=h, color=VehicleGraphics.BLUE)

        # Prepare a runtime canvas for rotation
        return self.obs_window.render(canvas=self.canvas_runtime, position=pos, heading=vehicle.heading_theta)

    @staticmethod
    def blit_rotate(
        surf: pygame.SurfaceType,
        image: pygame.SurfaceType,
        pos,
        angle: float,
    ) -> Tuple:
        """Many thanks to https://stackoverflow.com/a/54714144."""
        # calculate the axis aligned bounding box of the rotated image
        w, h = image.get_size()
        box = [pygame.math.Vector2(p) for p in [(0, 0), (w, 0), (w, -h), (0, -h)]]
        box_rotate = [p.rotate(angle) for p in box]
        min_box = (min(box_rotate, key=lambda p: p[0])[0], min(box_rotate, key=lambda p: p[1])[1])
        max_box = (max(box_rotate, key=lambda p: p[0])[0], max(box_rotate, key=lambda p: p[1])[1])

        # calculate the translation of the pivot
        origin_pos = w / 2, h / 2
        pivot = pygame.math.Vector2(origin_pos[0], -origin_pos[1])
        pivot_rotate = pivot.rotate(angle)
        pivot_move = pivot_rotate - pivot

        # calculate the upper left origin of the rotated image
        origin = (
            pos[0] - origin_pos[0] + min_box[0] - pivot_move[0], pos[1] - origin_pos[1] - max_box[1] + pivot_move[1]
        )
        # get a rotated image
        # rotated_image = pygame.transform.rotate(image, angle)
        rotated_image = pygame.transform.rotozoom(image, angle, 1.0)
        # rotate and blit the image
        surf.blit(rotated_image, origin)
        return origin

    def get_observation_window(self):
        return self.obs_window.get_observation_window()

    def get_screen_window(self):
        return self.obs_window.get_screen_window()

    @property
    def observation_space(self):
        shape = self.obs_shape + (self.num_stacks, )
        if self.rgb_clip:
            return gym.spaces.Box(-0.0, 1.0, shape=shape, dtype=np.float32)
        else:
            return gym.spaces.Box(0, 255, shape=shape, dtype=np.uint8)

    def observe(self, vehicle: BaseVehicle):
        self.render()
        surface = self.get_observation_window()
        img = self.pygame.surfarray.array3d(surface)
        if self.rgb_clip:
            img = img.astype(np.float32) / 255
        else:
            img = img.astype(np.uint8)
        return np.transpose(img, (1, 0, 2))



import numpy as np

from metadrive.component.lane.circular_lane import CircularLane
from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pg_space import Parameter, BlockParameterSpace
from metadrive.component.pgblock.bottleneck import Merge, Split
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.road_network import Road
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.component.vehicle_navigation_module.base_navigation import BaseNavigation
from metadrive.utils import clip, norm, get_np_random
from metadrive.utils.math import panda_vector
from metadrive.utils.pg.utils import ray_localization


class NodeNetworkNavigation(BaseNavigation):
    def __init__(
        self,
        show_navi_mark: bool = False,
        random_navi_mark_color=False,
        show_dest_mark=False,
        show_line_to_dest=False,
        panda_color=None,
        name=None,
        vehicle_config=None
    ):
        """
        This class define a helper for localizing vehicles and retrieving navigation information.
        It now only support from first block start to the end node, but can be extended easily.
        """
        super(NodeNetworkNavigation, self).__init__(
            show_navi_mark=show_navi_mark,
            random_navi_mark_color=random_navi_mark_color,
            show_dest_mark=show_dest_mark,
            show_line_to_dest=show_line_to_dest,
            panda_color=panda_color,
            name=name,
            vehicle_config=vehicle_config
        )
        self.final_road = None
        self.spawn_road = None
        self.current_road = None
        self.next_road = None

    def reset(self, vehicle):
        if not vehicle.config["need_navigation"]:
            return
        possible_lanes = ray_localization(vehicle.heading, vehicle.spawn_place, self.engine, use_heading_filter=False)
        possible_lane_indexes = [lane_index for lane, lane_index, dist in possible_lanes]

        if len(possible_lanes) == 0 and vehicle.config["spawn_lane_index"] is None:
            from metadrive.utils.error_class import NavigationError
            raise NavigationError("Can't find valid lane for navigation.")

        if vehicle.config["spawn_lane_index"] is not None and vehicle.config["spawn_lane_index"
                                                                             ] in possible_lane_indexes:
            idx = possible_lane_indexes.index(vehicle.config["spawn_lane_index"])
            lane, new_l_index = possible_lanes[idx][:-1]
        else:
            assert len(possible_lanes) > 0
            lane, new_l_index = possible_lanes[0][:-1]

        dest = vehicle.config["destination"]

        current_lane = lane
        destination = dest if dest is not None else None
        random_seed = self.engine.global_random_seed
        assert current_lane is not None, "spawn place is not on road!"
        super(NodeNetworkNavigation, self).reset(current_lane)
        assert self.map.road_network_type == NodeRoadNetwork, "This Navigation module only support NodeRoadNetwork type"
        destination = self.auto_assign_task(self.map, current_lane.index, destination, random_seed)
        self.set_route(current_lane.index, destination)

    @staticmethod
    def auto_assign_task(map, current_lane_index, final_road_node=None, random_seed=None):
        # TODO we will assign the route in the task manager in the future
        start_road_node = current_lane_index[0]
        if start_road_node is None:
            start_road_node = FirstPGBlock.NODE_1
        if final_road_node is None:
            current_road_negative = Road(*current_lane_index[:-1]).is_negative_road()
            # choose first block when born on negative road
            block = map.blocks[0] if current_road_negative else map.blocks[-1]
            sockets = block.get_socket_list()
            socket = get_np_random(random_seed).choice(sockets)
            while True:
                if not socket.is_socket_node(start_road_node) or len(sockets) == 1:
                    break
                else:
                    sockets.remove(socket)
                    if len(sockets) == 0:
                        raise ValueError("Can not set a destination!")
            # choose negative road end node when current road is negative road
            final_road_node = socket.negative_road.end_node if current_road_negative else socket.positive_road.end_node
        return final_road_node

    def set_route(self, current_lane_index: str, destination: str):
        """
        Find a shortest path from start road to end road
        :param current_lane_index: start road node
        :param destination: end road node or end lane index
        :return: None
        """
        self.spawn_road = current_lane_index[:-1]
        self.checkpoints = self.map.road_network.shortest_path(current_lane_index, destination)
        self._target_checkpoints_index = [0, 1]
        # update routing info
        if len(self.checkpoints) <= 2:
            self.checkpoints = [current_lane_index[0], current_lane_index[1]]
            self._target_checkpoints_index = [0, 0]
        assert len(self.checkpoints
                   ) >= 2, "Can not find a route from {} to {}".format(current_lane_index[0], destination)
        self.final_road = Road(self.checkpoints[-2], self.checkpoints[-1])
        final_lanes = self.final_road.get_lanes(self.map.road_network)
        self.final_lane = final_lanes[-1]
        self._navi_info.fill(0.0)
        target_road_1_start = self.checkpoints[0]
        target_road_1_end = self.checkpoints[1]
        self.current_ref_lanes = self.map.road_network.graph[target_road_1_start][target_road_1_end]

        for l in self.current_ref_lanes:
            assert l.index is not None, self.current_ref_lanes

        self.next_ref_lanes = self.map.road_network.graph[self.checkpoints[1]][self.checkpoints[2]
                                                                               ] if len(self.checkpoints) > 2 else None
        self.current_road = Road(target_road_1_start, target_road_1_end)
        self.next_road = Road(self.checkpoints[1], self.checkpoints[2]) if len(self.checkpoints) > 2 else None
        if self._dest_node_path is not None:
            ref_lane = final_lanes[0]
            later_middle = (float(self.get_current_lane_num()) / 2 - 0.5) * self.get_current_lane_width()
            check_point = ref_lane.position(ref_lane.length, later_middle)
            self._dest_node_path.setPos(panda_vector(check_point[0], check_point[1], 1.8))

    def update_localization(self, ego_vehicle):
        position = ego_vehicle.position
        lane, lane_index = self._update_current_lane(ego_vehicle)
        long, _ = lane.local_coordinates(position)
        need_update = self._update_target_checkpoints(lane_index, long)
        assert len(self.checkpoints) >= 2

        # target_road_1 is the road segment the vehicle is driving on.
        if need_update:
            target_road_1_start = self.checkpoints[self._target_checkpoints_index[0]]
            target_road_1_end = self.checkpoints[self._target_checkpoints_index[0] + 1]
            target_lanes_1 = self.map.road_network.graph[target_road_1_start][target_road_1_end]
            self.current_ref_lanes = target_lanes_1
            self.current_road = Road(target_road_1_start, target_road_1_end)

            # target_road_2 is next road segment the vehicle should drive on.
            target_road_2_start = self.checkpoints[self._target_checkpoints_index[1]]
            target_road_2_end = self.checkpoints[self._target_checkpoints_index[1] + 1]
            target_lanes_2 = self.map.road_network.graph[target_road_2_start][target_road_2_end]

            if target_road_1_start == target_road_2_start:
                # When we are in the final road segment that there is no further road to drive on
                self.next_road = None
                self.next_ref_lanes = None
            else:
                self.next_road = Road(target_road_2_start, target_road_2_end)
                self.next_ref_lanes = target_lanes_2

        self._navi_info.fill(0.0)
        half = self.CHECK_POINT_INFO_DIM
        self._navi_info[:half], lanes_heading1, checkpoint = self._get_info_for_checkpoint(
            lanes_id=0, ref_lane=self.current_ref_lanes[0], ego_vehicle=ego_vehicle
        )

        self._navi_info[half:], lanes_heading2, _ = self._get_info_for_checkpoint(
            lanes_id=1,
            ref_lane=self.next_ref_lanes[0] if self.next_ref_lanes is not None else self.current_ref_lanes[0],
            ego_vehicle=ego_vehicle
        )

        if self._show_navi_info:
            # Whether to visualize little boxes in the scene denoting the checkpoints
            pos_of_goal = checkpoint
            self._goal_node_path.setPos(panda_vector(pos_of_goal[0], pos_of_goal[1], 1.8))
            self._goal_node_path.setH(self._goal_node_path.getH() + 3)
            self.navi_arrow_dir = [lanes_heading1, lanes_heading2]
            dest_pos = self._dest_node_path.getPos()
            self._draw_line_to_dest(start_position=ego_vehicle.position, end_position=(dest_pos[0], dest_pos[1]))
            navi_pos = self._goal_node_path.getPos()
            self._draw_line_to_navi(start_position=ego_vehicle.position, end_position=(navi_pos[0], navi_pos[1]))

    def _update_target_checkpoints(self, ego_lane_index, ego_lane_longitude) -> bool:
        """
        Return should_update: True or False
        """
        if self._target_checkpoints_index[0] == self._target_checkpoints_index[1]:  # on last road
            return False

        # arrive to second checkpoint
        current_road_start_point = ego_lane_index[0]
        if current_road_start_point in self.checkpoints[self._target_checkpoints_index[1]:] \
                and ego_lane_longitude < self.CKPT_UPDATE_RANGE:
            if current_road_start_point not in self.checkpoints[self._target_checkpoints_index[1]:-1]:
                return False
            idx = self.checkpoints.index(current_road_start_point, self._target_checkpoints_index[1], -1)
            self._target_checkpoints_index = [idx]
            if idx + 1 == len(self.checkpoints) - 1:
                self._target_checkpoints_index.append(idx)
            else:
                self._target_checkpoints_index.append(idx + 1)
            return True
        return False

    def get_current_lateral_range(self, current_position, engine) -> float:
        """Return the maximum lateral distance from left to right."""
        # special process for special block
        try:
            current_block_id = self.current_road.block_ID()
        except AttributeError:
            return self.get_current_lane_width() * self.get_current_lane_num()
        if current_block_id == Split.ID or current_block_id == Merge.ID:
            left_lane = self.current_ref_lanes[0]
            assert isinstance(left_lane, StraightLane), "Reference lane should be straight lane here"
            long, lat = left_lane.local_coordinates(current_position)
            current_position = left_lane.position(long, -left_lane.width / 2)
            return self._ray_lateral_range(engine, current_position, self.current_ref_lanes[0].direction_lateral)
        else:
            return self.get_current_lane_width() * self.get_current_lane_num()

    def _get_current_lane(self, ego_vehicle):
        """
        Called in update_localization to find current lane information
        """
        possible_lanes, on_lane = ray_localization(
            ego_vehicle.heading, ego_vehicle.position, ego_vehicle.engine, return_on_lane=True
        )
        for lane, index, l_1_dist in possible_lanes:
            if lane in self.current_ref_lanes:
                return lane, index, on_lane
        nx_ckpt = self._target_checkpoints_index[-1]
        if nx_ckpt == self.checkpoints[-1] or self.next_ref_lanes is None:
            return (*possible_lanes[0][:-1], on_lane) if len(possible_lanes) > 0 else (None, None, on_lane)

        if self.map.road_network_type == NodeRoadNetwork:
            nx_nx_ckpt = nx_ckpt + 1
            next_ref_lanes = self.map.road_network.graph[self.checkpoints[nx_ckpt]][self.checkpoints[nx_nx_ckpt]]
        else:
            next_ref_lanes = self.next_ref_lanes
        for lane, index, l_1_dist in possible_lanes:
            if lane in next_ref_lanes:
                return lane, index, on_lane
        return (*possible_lanes[0][:-1], on_lane) if len(possible_lanes) > 0 else (None, None, on_lane)

    def _get_info_for_checkpoint(self, lanes_id, ref_lane, ego_vehicle):

        navi_information = []
        # Project the checkpoint position into the target vehicle's coordination, where
        # +x is the heading and +y is the right hand side.
        later_middle = (float(self.get_current_lane_num()) / 2 - 0.5) * self.get_current_lane_width()
        check_point = ref_lane.position(ref_lane.length, later_middle)
        dir_vec = check_point - ego_vehicle.position  # get the vector from center of vehicle to checkpoint
        dir_norm = norm(dir_vec[0], dir_vec[1])
        if dir_norm > self.NAVI_POINT_DIST:  # if the checkpoint is too far then crop the direction vector
            dir_vec = dir_vec / dir_norm * self.NAVI_POINT_DIST
        ckpt_in_heading, ckpt_in_rhs = ego_vehicle.convert_to_local_coordinates(
            dir_vec, 0.0
        )  # project to ego vehicle's coordination

        # Dim 1: the relative position of the checkpoint in the target vehicle's heading direction.
        navi_information.append(clip((ckpt_in_heading / self.NAVI_POINT_DIST + 1) / 2, 0.0, 1.0))

        # Dim 2: the relative position of the checkpoint in the target vehicle's right hand side direction.
        navi_information.append(clip((ckpt_in_rhs / self.NAVI_POINT_DIST + 1) / 2, 0.0, 1.0))

        if lanes_id == 0:
            lanes_heading = ref_lane.heading_theta_at(ref_lane.local_coordinates(ego_vehicle.position)[0])
        else:
            lanes_heading = ref_lane.heading_theta_at(min(self.PRE_NOTIFY_DIST, ref_lane.length))

        # Try to include the current lane's information into the navigation information
        bendradius = 0.0
        dir = 0.0
        angle = 0.0
        if isinstance(ref_lane, CircularLane):
            bendradius = ref_lane.radius / (
                BlockParameterSpace.CURVE[Parameter.radius].max +
                self.get_current_lane_num() * self.get_current_lane_width()
            )
            dir = -ref_lane.direction
            angle = ref_lane.angle

        # Dim 3: The bending radius of current lane
        navi_information.append(clip(bendradius, 0.0, 1.0))

        # Dim 4: The bending direction of current lane (+1 for clockwise, -1 for counterclockwise)
        navi_information.append(clip((dir + 1) / 2, 0.0, 1.0))

        # Dim 5: The angular difference between the heading in lane ending position and
        # the heading in lane starting position
        navi_information.append(
            clip((np.rad2deg(angle) / BlockParameterSpace.CURVE[Parameter.angle].max + 1) / 2, 0.0, 1.0)
        )
        return navi_information, lanes_heading, check_point

    def _update_current_lane(self, ego_vehicle):
        lane, lane_index, on_lane = self._get_current_lane(ego_vehicle)
        ego_vehicle.on_lane = on_lane
        if lane is None:
            lane, lane_index = ego_vehicle.lane, ego_vehicle.lane_index
            if self.FORCE_CALCULATE:
                lane_index, _ = self.map.road_network.get_closest_lane_index(ego_vehicle.position)
                lane = self.map.road_network.get_lane(lane_index)
        self._current_lane = lane
        assert lane_index == lane.index, "lane index mismatch!"
        return lane, lane_index

    def get_state(self):
        final_road = self.final_road
        return {"spawn_road": self.spawn_road, "destination": (final_road.start_node, final_road.end_node)}



__all__ = ['BaseLoader', 'SafeLoader', 'Loader']

from reader import *
from scanner import *
from parser import *
from composer import *
from constructor import *
from resolver import *


class BaseLoader(Reader, Scanner, Parser, Composer, BaseConstructor, BaseResolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        BaseConstructor.__init__(self)
        BaseResolver.__init__(self)


class SafeLoader(Reader, Scanner, Parser, Composer, SafeConstructor, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)


class Loader(Reader, Scanner, Parser, Composer, Constructor, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        Constructor.__init__(self)
        Resolver.__init__(self)



import numpy as np

from metadrive import MetaDriveEnv
from metadrive.component.vehicle.vehicle_type import DefaultVehicle
from metadrive.envs.base_env import BASE_DEFAULT_CONFIG
from metadrive.envs.metadrive_env import METADRIVE_DEFAULT_CONFIG
from metadrive.utils import Config
from metadrive.utils.coordinates_shift import metadrive_vector


def _assert_vehicle(vehicle):
    pos = vehicle.position
    assert -200 < pos[0] < 200
    assert -200 < pos[1] < 200
    speed = vehicle.speed_km_h
    assert 0 <= speed <= 120
    velocity_direction = vehicle.heading
    np.testing.assert_almost_equal(abs(np.linalg.norm(velocity_direction)), 1.0)
    current_road = vehicle.navigation.current_road
    np.testing.assert_almost_equal(vehicle.heading_diff(vehicle.lane), 0.5, decimal=3)


def _get_heading_deg(heading):
    return (heading * 180 / np.pi)


def test_base_vehicle():
    env = MetaDriveEnv()
    try:
        env.reset()
        engine = env.engine
        map = env.current_map

        # v_config = BaseVehicle.get_vehicle_config(dict())
        v_config = Config(BASE_DEFAULT_CONFIG["vehicle_config"]).update(METADRIVE_DEFAULT_CONFIG["vehicle_config"])
        v = engine.spawn_object(DefaultVehicle, vehicle_config=v_config, random_seed=0)

        v.add_navigation()
        v.add_navigation()
        v.navigation.set_force_calculate_lane_index(True)
        v.reset_navigation()

        for heading in [-1.0, 0.0, 1.0]:
            for pos in [[0., 0.], [-100., -100.], [100., 100.]]:
                # v.reset(position=pos, heading=heading)
                # np.testing.assert_almost_equal(_get_heading_deg(v.heading_theta), heading, decimal=3)
                #
                # v_pos = v.position
                # # v_pos[1] = -v_pos[1], this position is converted to pg_position in reset() now
                # np.testing.assert_almost_equal(v_pos, pos)

                v.set_position(pos)
                v_pos = v.position
                np.testing.assert_almost_equal(v_pos, pos)

                v.after_step()
        v.reset(position=np.array([10, 0]))
        for a_x in [-1, 0, 0.5, 1]:
            for a_y in [-1, 0, 0.5, 1]:
                v.before_step([a_x, a_y])
                v._set_action([a_x, a_y])
                _assert_vehicle(v)
                v._set_incremental_action([a_x, a_y])
                _assert_vehicle(v)
                state = v.get_state()
                v.set_state(state)
                assert _get_heading_deg(v.heading_theta) == _get_heading_deg(state["heading_theta"])
                np.testing.assert_almost_equal(v.position, metadrive_vector(state["position"]))
                v.convert_to_local_coordinates([a_x, a_y], 0.0)

        _nan_speed(env)

        v.reset(position=np.array([10, 0]))
        v.set_heading_theta(np.pi / 2)
        bounding_box = np.asarray(v.bounding_box)
        l = v.LENGTH
        w = v.WIDTH
        points = np.asarray([
            [10 - w / 2, l / 2],
            [10 + w / 2, l / 2],
            [10 + w / 2, -l / 2],
            [10 - w / 2, -l / 2],
        ])
        np.testing.assert_almost_equal(bounding_box, points)

        v.destroy()
        del v
    finally:
        env.close()


def _nan_speed(pg_env):
    steering = [-np.nan, -1, 0, 1, np.nan]
    acc_brake = [-np.nan, -1, 0, 1, np.nan]
    pg_env.reset()
    for s in steering:
        for a in acc_brake:
            pg_env.step([s, a])


if __name__ == '__main__':
    # pytest.main(["-sv", "test_ego_vehicle.py"])
    test_base_vehicle()



# Author: Epihaius
# Date: 2019-06-21
#
# This package contains classes to procedurally create 3D primitives such as
# boxes, spheres and cylinders.
# see: https://github.com/Epihaius/procedural_panda3d_model_primitives for more details

from .box import BoxMaker
from .cylinder import CylinderMaker
from .cone import ConeMaker
from .sphere import SphereMaker
from .torus import TorusMaker



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from metadrive.third_party.kitsunetsuki.base.armature import get_armature
from metadrive.third_party.kitsunetsuki.base.matrices import get_object_matrix
from metadrive.third_party.kitsunetsuki.base.mesh import obj2mesh
from metadrive.third_party.kitsunetsuki.base.objects import apply_modifiers, is_collision

from . import spec


class GeomMixin(object):
    def _get_joints(self, gltf_node):
        results = {}

        for i, child_id in enumerate(gltf_node['joints']):
            child = self._root['nodes'][child_id]
            results[child['name']] = i

        return results

    def _make_primitive(self, gltf_mesh, mesh):
        gltf_primitive = {
            'attributes': {},
            'material': 0,
            # 'mode': 4,
            'extras': {
                'highest_index': -1,
                'targetNames': gltf_mesh['extras']['targetNames'],
            },
        }

        channel = self._buffer.add_channel(
            {
                # 'componentType': spec.TYPE_UNSIGNED_SHORT,
                'componentType': spec.TYPE_UNSIGNED_INT,
                'type': 'SCALAR',
                'extras': {
                    'reference': 'indices',
                },
            }
        )
        gltf_primitive['indices'] = channel['bufferView']

        if gltf_mesh['primitives'] and not self._split_primitives:
            gltf_primitive['attributes'] = gltf_mesh['primitives'][0]['attributes']
            if 'targets' in gltf_mesh['primitives'][0]:
                gltf_primitive['targets'] = gltf_mesh['primitives'][0]['targets']

        else:
            channel = self._buffer.add_channel(
                {
                    'componentType': spec.TYPE_FLOAT,
                    'type': 'VEC3',
                    'extras': {
                        'reference': 'NORMAL',
                    },
                }
            )
            gltf_primitive['attributes']['NORMAL'] = channel['bufferView']

            channel = self._buffer.add_channel(
                {
                    'componentType': spec.TYPE_FLOAT,
                    'type': 'VEC3',
                    'extras': {
                        'reference': 'POSITION',
                    },
                }
            )
            gltf_primitive['attributes']['POSITION'] = channel['bufferView']

            for sk_id, sk_name in enumerate(gltf_primitive['extras']['targetNames']):
                gltf_target = {
                    # '_extras': {
                    #     'name': sk_name,
                    # },
                }

                channel = self._buffer.add_channel(
                    {
                        'componentType': spec.TYPE_FLOAT,
                        'type': 'VEC3',
                        'extras': {
                            'reference': 'POSITION',
                            'target': sk_name,
                        },
                    }
                )
                gltf_target['POSITION'] = channel['bufferView']

                if 'targets' not in gltf_primitive:
                    gltf_primitive['targets'] = []
                gltf_primitive['targets'].append(gltf_target)

        return gltf_primitive

    def make_geom(self, gltf_node, gltf_mesh, obj, can_merge=False):
        triangulate = True
        if self._geom_scale != 1:
            scale = obj.scale
            obj.scale.x, obj.scale.y, obj.scale.z = [self._geom_scale] * 3
            apply_modifiers(obj, triangulate=triangulate, apply_scale=True)
            obj.scale = scale
        else:
            apply_modifiers(obj, triangulate=triangulate)
        mesh = obj2mesh(obj, triangulate=triangulate)

        # setup shape key names for the primitives
        if mesh.shape_keys:
            for sk_name in sorted(mesh.shape_keys.key_blocks.keys()):
                if sk_name.lower() == 'basis':
                    continue

                gltf_mesh['extras']['targetNames'].append(sk_name)

        # get or create materials and textures
        gltf_materials = {}
        if not self._no_materials and not is_collision(obj):
            for material in mesh.materials.values():
                # empty material slot
                if not material:
                    continue

                # material
                for i, child in enumerate(self._root['materials']):  # existing material
                    if child['name'] == material.name:
                        gltf_materials[material.name] = i
                        break
                else:  # new material
                    gltf_material = self.make_material(material)
                    self._root['materials'].append(gltf_material)

                    gltf_materials[material.name] = len(self._root['materials']) - 1

                # textures
                if not self._no_textures:
                    for type_, gltf_sampler, gltf_image in self.make_textures(material):
                        tname = gltf_image['name']
                        for i, child in enumerate(self._root['images']):  # existing texture
                            if child['name'] == tname:
                                texid = i
                                break
                        else:  # new texture
                            self._root['samplers'].append(gltf_sampler)
                            self._root['images'].append(gltf_image)

                            gltf_texture = {
                                'sampler': len(self._root['samplers']) - 1,
                                'source': len(self._root['images']) - 1,
                            }
                            self._root['textures'].append(gltf_texture)
                            texid = len(self._root['textures']) - 1

                        matid = gltf_materials[material.name]
                        if type(type_) == tuple and len(type_) == 2:
                            type_l1, type_l2 = type_
                            self._root['materials'][matid][type_l1][type_l2] = {
                                'index': texid,
                                'texCoord': 0,
                            }
                        else:
                            self._root['materials'][matid][type_] = {
                                'index': texid,
                                'texCoord': 0,
                            }

        # get primitives
        gltf_primitives = {}
        gltf_primitive_indices = {}  # splitted vertex buffers
        gltf_mesh_vertices_index = -1  # reusable vertex buffer
        if can_merge:
            for i, gltf_primitive in enumerate(gltf_mesh['primitives']):
                mname = None
                if 'material' in gltf_primitive:
                    matid = gltf_primitive['material']
                    mname = self._root['materials'][matid]['name']
                gltf_primitives[mname] = gltf_primitive
                gltf_primitive_indices[mname] = gltf_primitive['extras']['highest_index']

        gltf_vertices = {}

        # get armature and joints
        armature = get_armature(obj)
        max_joints = 0  # get max joints per vertex
        gltf_joints = {}
        if armature:
            # max_joints = 1
            # for polygon in mesh.polygons:
            #     for vertex_id in polygon.vertices:
            #         vertex = mesh.vertices[vertex_id]
            #         joints = 0
            #         for vertex_group in vertex.groups:
            #             obj_vertex_group = obj.vertex_groups[vertex_group.group]
            #             if vertex_group.weight > 0:
            #                 joints += 1
            #         max_joints = max(max_joints, joints)

            if 'skin' in gltf_node:
                gltf_skin = self._root['skins'][gltf_node['skin']]

                # for i, child in enumerate(self._root['skins']):
                #     if child['name'] == armature.name:
                #         gltf_joints = self._get_joints(child)
                #         break
                gltf_joints = self._get_joints(gltf_skin)

        # get max joint layers (4 bones per layer)
        # max_joint_layers = math.ceil(max_joints / 4)

        # panda3d-gltf is limited to 1 single layer only (up to 4 bones)
        max_joint_layers = 1

        sharp_vertices = self.get_sharp_vertices(mesh)
        uv_tb = self.get_tangent_bitangent(mesh)
        obj_matrix = self._transform(get_object_matrix(obj, armature=armature))

        for polygon in mesh.polygons:
            # <-- polygon
            material = None
            mname = None
            if not self._no_materials:
                try:
                    material = mesh.materials[polygon.material_index]
                    if material:
                        mname = material.name
                except IndexError:
                    pass

            # get or create primitive
            if mname in gltf_primitives:
                gltf_primitive = gltf_primitives[mname]
            else:
                gltf_primitive = self._make_primitive(gltf_mesh, mesh)
                gltf_primitives[mname] = gltf_primitive
                gltf_primitive_indices[mname] = -1
                gltf_mesh['primitives'].append(gltf_primitive)

            # set material
            if material and not self._no_materials and not is_collision(obj):
                if material.name in gltf_materials:
                    gltf_primitive['material'] = gltf_materials[mname]

            # vertices
            for i, vertex_id in enumerate(polygon.vertices):
                # i is vertex counter inside a polygon
                # (0, 1, 2) for triangle
                # vertex_id is reusable id,
                # because multiple polygons can share the same vertices

                loop_id = polygon.loop_indices[i]

                # <-- vertex
                vertex = mesh.vertices[vertex_id]
                use_smooth = (
                    polygon.use_smooth and
                    # vertex_id not in sharp_vertices and
                    not is_collision(obj)
                )

                # try to reuse shared vertices
                if mname not in gltf_vertices:
                    gltf_vertices[mname] = {}
                if (polygon.use_smooth and vertex_id in gltf_vertices[mname] and not is_collision(obj)):
                    shared = False
                    for gltf_vertex_index, gltf_vertex_uv, gltf_vertex_normal in gltf_vertices[mname][vertex_id]:
                        if self.can_share_vertex(mesh, vertex, loop_id, gltf_vertex_uv, gltf_vertex_normal):
                            self._buffer.write(gltf_primitive['indices'], gltf_vertex_index)
                            shared = True
                            break
                    if shared:
                        continue

                # make new vertex data
                can_merge_vertices = can_merge
                if armature:
                    can_merge_vertices = True
                elif is_collision(obj):
                    can_merge_vertices = False
                self.make_vertex(
                    obj_matrix,
                    gltf_primitive,
                    mesh,
                    polygon,
                    vertex,
                    vertex_id,
                    loop_id,
                    use_smooth=use_smooth,
                    can_merge=can_merge_vertices
                )

                # uv layers, active first
                active_uv = 0, 0
                if not is_collision(obj):
                    uv_layers = sorted(mesh.uv_layers.items(), key=lambda x: not x[1].active)
                    for uv_id, (uv_name, uv_layer) in enumerate(uv_layers):
                        # <-- vertex uv
                        uv_loop = uv_layer.data[loop_id]

                        # not active layer and extra UV disabled
                        if not uv_layer.active and self._no_extra_uv:
                            continue

                        u, v = uv_loop.uv.to_2d()
                        if uv_layer.active:
                            active_uv = u, v
                        self._write_uv(gltf_primitive, uv_id, u, v)
                        if uv_name in uv_tb and uv_layer.active:
                            self._write_tbs(
                                obj_matrix, gltf_primitive, *uv_tb[uv_name][loop_id], can_merge=can_merge_vertices
                            )
                        # vertex uv -->

                # generate new ID, add vertex and save last ID
                gltf_primitive_indices[mname] += 1
                gltf_mesh_vertices_index += 1
                if self._split_primitives:
                    idx = gltf_primitive_indices[mname]
                else:
                    idx = gltf_mesh_vertices_index
                self._buffer.write(gltf_primitive['indices'], idx)
                gltf_primitive['extras']['highest_index'] = gltf_primitive_indices[mname]

                # save vertex data for sharing
                if vertex_id not in gltf_vertices[mname]:
                    gltf_vertices[mname][vertex_id] = []
                gltf_vertices[mname][vertex_id].append(
                    (
                        idx,
                        active_uv,
                        mesh.loops[loop_id].normal if use_smooth else polygon.normal,
                    )
                )

                # attach joints to vertex
                if gltf_joints:
                    joints_weights = []
                    vertex_groups = reversed(sorted(vertex.groups, key=lambda vg: vg.weight))
                    for vertex_group in vertex_groups:
                        obj_vertex_group = obj.vertex_groups[vertex_group.group]

                        # no bones with vertex group's name
                        if obj_vertex_group.name not in gltf_joints:
                            continue

                        # weight is zero
                        if vertex_group.weight <= 0:
                            continue

                        joint_id = gltf_joints[obj_vertex_group.name]
                        joints_weights.append([joint_id, vertex_group.weight])

                    # objects reparented to bone instead of entire armature
                    if obj.parent_type == 'BONE' and obj.parent_bone in gltf_joints:
                        joint_id = gltf_joints[obj.parent_bone]
                        joints_weights.append([joint_id, 1])

                    # padding
                    while ((len(joints_weights) % 4 != 0) or (len(joints_weights) < max_joint_layers * 4)):
                        joints_weights.append([0, 0])

                    # limit by max joints
                    joints_weights = joints_weights[:max_joint_layers * 4]

                    imax = -1
                    wmax = 0
                    for j, (joint, weight) in enumerate(joints_weights):
                        if weight > wmax:
                            imax = j
                            wmax = weight
                    # if self._norm_weights and imax >= 0:
                    #     joints_weights[imax][1] += 1 - sum(list(zip(*joints_weights))[1])
                    if self._norm_weights:
                        weights = list(zip(*joints_weights))[1]
                        weights_sum = sum(weights)
                        for j in range(len(joints_weights)):
                            joints_weights[j][1] *= 1 / weights_sum

                    # group by 4 joint-weight pairs
                    joints_weights_groups = []
                    for j in range(len(joints_weights) // 4):
                        group = joints_weights[j * 4:j * 4 + 4]
                        joints_weights_groups.append(group)

                    self._write_joints_weights(gltf_primitive, len(tuple(gltf_joints.keys())), joints_weights_groups)

                # vertex -->
            # polygon -->



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import json
import math
import itertools

from panda3d.core import CS_zup_right, LMatrix4d
from panda3d.egg import EggComment, EggData, EggGroup, EggPolygon, EggTransform

from metadrive.third_party.kitsunetsuki.base.armature import get_armature
from metadrive.third_party.kitsunetsuki.base.collections import get_object_collection
from metadrive.third_party.kitsunetsuki.base.matrices import get_object_matrix, get_bone_matrix
from metadrive.third_party.kitsunetsuki.base.objects import (is_collision, get_object_properties, set_active_object)

from metadrive.third_party.kitsunetsuki.exporter.base import Exporter

from .animation import AnimationMixin
from .geom import GeomMixin
from .material import MaterialMixin
from .texture import TextureMixin
from .vertex import VertexMixin


def matrix_to_panda(matrix):
    return LMatrix4d(*itertools.chain(*map(tuple, matrix.col)))


class EggExporter(AnimationMixin, GeomMixin, MaterialMixin, TextureMixin, VertexMixin, Exporter):
    """
    BLEND to EGG converter.
    """
    def __init__(self, args):
        super().__init__(args)
        self._output = args.output or args.inputs[0].replace('.blend', '.egg')

    def make_root_node(self):
        egg_root = EggData()
        egg_root.set_coordinate_system(CS_zup_right)  # Z-up

        egg_comment = EggComment(
            '', 'KITSUNETSUKI Asset Tools by kitsune.ONE - '
            'https://github.com/kitsune-ONE-team/KITSUNETSUKI-Asset-Tools'
        )
        egg_root.add_child(egg_comment)

        return egg_root

    def _setup_node(self, node, obj=None, can_merge=False):
        if obj is None:
            return

        armature = get_armature(obj)
        obj_matrix = get_object_matrix(obj, armature=armature)

        # get custom object properties
        obj_props = get_object_properties(obj)

        if not can_merge and not armature:
            node.add_matrix4(matrix_to_panda(obj_matrix))

        if obj_props.get('type') == 'Portal':
            node.set_portal_flag(True)

        # setup collisions
        if not can_merge and is_collision(obj):
            # collision name
            node.set_collision_name(obj.name)

            # collision solid type
            shape = {
                'BOX': EggGroup.CST_box,
                'SPHERE': EggGroup.CST_sphere,
                'CAPSULE': EggGroup.CST_tube,
                'MESH': EggGroup.CST_polyset,
            }.get(obj.rigid_body.collision_shape, EggGroup.CST_polyset)

            # custom shape
            if obj.rigid_body.collision_shape == 'CONVEX_HULL':
                # trying to guess the best shape
                polygons = list(filter(lambda x: isinstance(x, EggPolygon), node.get_children()))
                if len(polygons) == 1 and polygons[0].is_planar():
                    # shape = EggGroup.CST_plane  # <- is it infinite?
                    shape = EggGroup.CST_polygon
            node.set_cs_type(shape)

            # collision flags
            # inherit collision by children nodes?
            flags = EggGroup.CF_descend
            if (not obj.collision or not obj.collision.use):
                # don't actually collide (ghost)
                flags |= EggGroup.CF_intangible
            node.set_collide_flags(flags)

            if self._set_origin:
                obj.select_set(state=True)
                set_active_object(obj)
                x1, y1, z1 = obj.location
                # set origin to the center of bounds
                bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='BOUNDS')
                x2, y2, z2 = obj.location
                node.set_tag('origin', json.dumps([x2 - x1, y2 - y1, z2 - z1]))

        # setup custom properties with tags
        for k, v in obj_props.items():
            if node.get_tag(k):  # tag exists
                tag = node.get_tag(k)

            if type(v) in (tuple, list, dict):
                tag = json.dumps(v)
            else:
                tag = '{}'.format(v)
            node.set_tag(k, tag)

        # if can_merge and 'type' not in obj_props:
        if can_merge:
            node.set_tag('type', 'Merged')

    def make_empty(self, parent_node, obj):
        egg_group = EggGroup(obj.name)

        self._setup_node(egg_group, obj)
        parent_node.add_child(egg_group)

        return egg_group

    def make_armature(self, parent_node, armature):
        egg_group = EggGroup(armature.name)
        egg_group.set_dart_type(EggGroup.DT_structured)

        egg_joints = {}

        for bone_name, bone in armature.data.bones.items():
            bone_matrix = get_bone_matrix(bone, armature)

            egg_transform = EggTransform()
            egg_transform.add_matrix4(matrix_to_panda(bone_matrix))

            egg_joint = EggGroup(bone_name)
            egg_joint.set_group_type(EggGroup.GT_joint)
            egg_joint.add_matrix4(matrix_to_panda(bone_matrix))
            egg_joint.set_default_pose(egg_transform)

            if bone.parent:
                egg_joints[bone.parent.name].add_child(egg_joint)
            else:  # root bone
                egg_group.add_child(egg_joint)

            egg_joints[bone_name] = egg_joint

        self._setup_node(egg_group, armature)
        parent_node.add_child(egg_group)

        return egg_group

    def make_mesh(self, parent_node, obj):
        egg_group = None

        # merged nodes
        if self.can_merge(obj):
            collection = get_object_collection(obj)

            for child in parent_node.get_children():
                if (isinstance(child, EggGroup) and child.get_name() == collection.name):
                    egg_group = child
                    break
            else:
                egg_group = EggGroup(collection.name)
                self._setup_node(egg_group, obj, can_merge=True)
                parent_node.add_child(egg_group)

            self.make_geom(egg_group, obj, can_merge=True)

        # separate nodes
        if not self.can_merge(obj) or self._keep:
            egg_group = EggGroup(obj.name)
            self._setup_node(egg_group, obj, can_merge=False)
            parent_node.add_child(egg_group)

            self.make_geom(egg_group, obj, can_merge=False)

        return egg_group

    def make_light(self, parent_node, obj):
        LIGHT_TYPES = {
            'POINT': 'PointLight',
            'SPOT': 'SpotLight',
        }

        egg_group = EggGroup(obj.name)
        egg_group.set_tag('type', 'Light')
        egg_group.set_tag('light', LIGHT_TYPES[obj.data.type])

        egg_group.set_tag('color', json.dumps(tuple(obj.data.color)))
        egg_group.set_tag('scale', json.dumps(tuple(obj.scale)))
        egg_group.set_tag('energy', '{:.3f}'.format(obj.data.energy))
        egg_group.set_tag('far', '{:.3f}'.format(obj.data.shadow_soft_size))

        if obj.data.type == 'SPOT':
            egg_group.set_tag('fov', '{:.3f}'.format(math.degrees(obj.data.spot_size)))

        self._setup_node(egg_group, obj)
        parent_node.add_child(egg_group)

        return egg_group



import math
import numpy as np

from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import CreateAdverseRoad, CreateRoadFrom, create_bend_straight
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class Roundabout(PGBlock):
    """
    roundabout class, the example is the same as Intersection
    """
    ID = "O"
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.ROUNDABOUT)
    SOCKET_NUM = 3
    RADIUS_IN = 20
    ANGLE = 60
    EXIT_PART_LENGTH = 35

    def __init__(self, *args, **kwargs):
        super(Roundabout, self).__init__(*args, **kwargs)
        self.intermediate_spawn_places = []

    def _try_plug_into_previous_block(self) -> bool:
        self.intermediate_spawn_places = []
        para = self.get_config(copy=False)
        no_cross = True
        attach_road = self.pre_block_socket.positive_road
        for i in range(4):
            exit_road, success = self._create_circular_part(
                attach_road, i, para[Parameter.radius_exit], para[Parameter.radius_inner], para[Parameter.angle]
            )
            no_cross = no_cross and success
            if i < 3:
                no_cross = CreateAdverseRoad(
                    exit_road,
                    self.block_network,
                    self._global_network,
                    ignore_intersection_checking=self.ignore_intersection_checking
                ) and no_cross
                attach_road = -exit_road
        self.add_respawn_roads([socket.negative_road for socket in self.get_socket_list()])
        return no_cross

    def _create_circular_part(self, road: Road, part_idx: int, radius_exit: float, radius_inner: float,
                              angle: float) -> (str, str, StraightLane, bool):
        """
        Create a part of roundabout according to a straight road
        """
        none_cross = True
        self.set_part_idx(part_idx)
        radius_big = (self.positive_lane_num * 2 - 1) * self.lane_width + radius_inner

        # circular part 0
        segment_start_node = road.end_node
        segment_end_node = self.add_road_node()
        segment_road = Road(segment_start_node, segment_end_node)
        lanes = road.get_lanes(self._global_network) if part_idx == 0 else road.get_lanes(self.block_network)
        right_lane = lanes[-1]
        bend, straight = create_bend_straight(
            right_lane, 10, radius_exit, np.deg2rad(angle), True, self.lane_width, (PGLineType.BROKEN, PGLineType.SIDE)
        )
        ignore_last_2_part_start = self.road_node((part_idx + 3) % 4, 0)
        ignore_last_2_part_end = self.road_node((part_idx + 3) % 4, 0)
        none_cross = CreateRoadFrom(
            bend,
            self.positive_lane_num,
            segment_road,
            self.block_network,
            self._global_network,
            ignore_start=ignore_last_2_part_start,
            ignore_end=ignore_last_2_part_end,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and none_cross

        # set circular part 0 visualization
        for k, lane in enumerate(segment_road.get_lanes(self.block_network)):
            if k == self.positive_lane_num - 1:
                lane.line_types = [PGLineType.NONE, PGLineType.SIDE]
            else:
                lane.line_types = [PGLineType.NONE, PGLineType.NONE]

        # circular part 1
        tool_lane_start = straight.position(-5, 0)
        tool_lane_end = straight.position(0, 0)
        tool_lane = StraightLane(tool_lane_start, tool_lane_end)

        bend, straight_to_next_iter_part = create_bend_straight(
            tool_lane, 10, radius_big, np.deg2rad(2 * angle - 90), False, self.lane_width,
            (PGLineType.BROKEN, PGLineType.SIDE)
        )

        segment_start_node = segment_end_node
        segment_end_node = self.add_road_node()
        segment_road = Road(segment_start_node, segment_end_node)

        none_cross = CreateRoadFrom(
            bend,
            self.positive_lane_num,
            segment_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and none_cross

        self.intermediate_spawn_places.append(segment_road.get_lanes(self.block_network))

        # circular part 2 and exit straight part
        length = self.EXIT_PART_LENGTH
        tool_lane_start = straight_to_next_iter_part.position(-5, 0)
        tool_lane_end = straight_to_next_iter_part.position(0, 0)
        tool_lane = StraightLane(tool_lane_start, tool_lane_end)

        bend, straight = create_bend_straight(
            tool_lane, length, radius_exit, np.deg2rad(angle), True, self.lane_width,
            (PGLineType.BROKEN, PGLineType.SIDE)
        )

        segment_start_node = segment_end_node
        segment_end_node = self.add_road_node() if part_idx < 3 else self.pre_block_socket.negative_road.start_node
        segment_road = Road(segment_start_node, segment_end_node)

        none_cross = CreateRoadFrom(
            bend,
            self.positive_lane_num,
            segment_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and none_cross

        # set circular part 2 (curve) visualization
        for k, lane in enumerate(segment_road.get_lanes(self.block_network)):
            if k == self.positive_lane_num - 1:
                lane.line_types = [PGLineType.NONE, PGLineType.SIDE]
            else:
                lane.line_types = [PGLineType.NONE, PGLineType.NONE]

        exit_start = segment_end_node
        exit_end = self.add_road_node()
        segment_road = Road(exit_start, exit_end)
        if part_idx < 3:
            none_cross = CreateRoadFrom(
                straight,
                self.positive_lane_num,
                segment_road,
                self.block_network,
                self._global_network,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and none_cross
            self.add_sockets(self.create_socket_from_positive_road(segment_road))

        #  add circular part 3 at last
        segment_start = self.road_node(part_idx, 1)
        segment_end = self.road_node((part_idx + 1) % 4, 0)
        segment_road = Road(segment_start, segment_end)
        tool_lane_start = straight_to_next_iter_part.position(-6, 0)
        tool_lane_end = straight_to_next_iter_part.position(0, 0)
        tool_lane = StraightLane(tool_lane_start, tool_lane_end)

        beneath = (self.positive_lane_num * 2 - 1) * self.lane_width / 2 + radius_exit
        cos = math.cos(np.deg2rad(angle))
        radius_this_seg = beneath / cos - radius_exit

        bend, _ = create_bend_straight(
            tool_lane, 5, radius_this_seg, np.deg2rad(180 - 2 * angle), False, self.lane_width,
            (PGLineType.BROKEN, PGLineType.SIDE)
        )
        CreateRoadFrom(
            bend,
            self.positive_lane_num,
            segment_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        # set circular part 2 visualization
        for k, lane in enumerate(segment_road.get_lanes(self.block_network)):
            if k == 0:
                if self.positive_lane_num > 1:
                    lane.line_types = [PGLineType.CONTINUOUS, PGLineType.BROKEN]
                else:
                    lane.line_types = [PGLineType.CONTINUOUS, PGLineType.NONE]
            else:
                lane.line_types = [PGLineType.BROKEN, PGLineType.BROKEN]

        return Road(exit_start, exit_end), none_cross

    def get_socket(self, index: int) -> PGBlockSocket:
        socket = super(Roundabout, self).get_socket(index)
        if socket.negative_road in self.get_respawn_roads():
            self._respawn_roads.remove(socket.negative_road)
        return socket

    def get_intermediate_spawn_lanes(self):
        """Filter out half of the vehicles."""
        return self.get_respawn_lanes() + self.intermediate_spawn_places



import numpy as np

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import recursive_equal, norm


def test_seeding():
    env = MetaDriveEnv({"num_scenarios": 1000})
    try:
        env.reset()
        env.seed(999)
        # assert env.engine is None
        assert env.current_seed == 999
        env.reset(seed=992)
        assert env.current_seed == 992
        # assert env.engine is not None
    finally:
        env.close()


def test_map_random_seeding():
    cfg_1 = {
        "num_scenarios": 1,
        "start_seed": 5,
        "random_lane_width": True,
        "random_lane_num": True,
    }
    cfg_2 = {
        "num_scenarios": 10,
        "start_seed": 5,
        "random_lane_width": True,
        "random_lane_num": True,
    }
    cfg_3 = {
        "num_scenarios": 100,
        "start_seed": 5,
        "random_lane_width": True,
        "random_lane_num": True,
    }
    cfg_4 = {
        "num_scenarios": 10,
        "start_seed": 0,
        "random_lane_width": True,
        "random_lane_num": True,
    }
    cfg_5 = {
        "num_scenarios": 3,
        "start_seed": 3,
        "random_lane_width": True,
        "random_lane_num": True,
    }
    from metadrive.component.pgblock.first_block import FirstPGBlock
    map_configs = []
    lane_width = []
    lane_num = []
    for cfg in [cfg_1, cfg_2, cfg_3, cfg_4, cfg_5]:
        env = MetaDriveEnv(cfg)
        try:
            env.reset()
            env.reset()
            env.reset(seed=5)
            map_configs.append(env.current_map.get_meta_data())
            lane_num.append(len(env.current_map.road_network.graph[FirstPGBlock.NODE_1][FirstPGBlock.NODE_2]))
            lane_width.append(
                env.current_map.road_network.graph[FirstPGBlock.NODE_1][FirstPGBlock.NODE_2][0].width_at(0)
            )
        finally:
            env.close()
    for idx, map_cfg in enumerate(map_configs[:-1]):
        nxt_map_cfg = map_configs[idx + 1]
        ret = recursive_equal(map_cfg, nxt_map_cfg)
        assert ret, "Error"
    assert np.std(lane_width) < 0.01 and np.std(lane_num) < 0.01, "random engine error"


def test_fixed_traffic():
    env = MetaDriveEnv({
        "random_traffic": False,
        "traffic_mode": "respawn",
        #  "use_render": True
    })
    try:
        last_pos = None
        for i in range(20):
            obs, _ = env.reset()
            assert env.engine.traffic_manager.random_seed == env.current_seed
            new_pos = [v.position for v in env.engine.traffic_manager.vehicles]
            if last_pos is not None and len(new_pos) == len(last_pos):
                assert sum(
                    [norm(lastp[0] - newp[0], lastp[1] - newp[1]) <= 1e-3 for lastp, newp in zip(last_pos, new_pos)]
                ), [(lastp, newp) for lastp, newp in zip(last_pos, new_pos)]
            last_pos = new_pos
    finally:
        env.close()


def test_random_traffic():
    env = MetaDriveEnv(
        {
            "random_traffic": True,
            "traffic_mode": "respawn",
            "traffic_density": 0.3,
            "start_seed": 5,

            #  "use_render": True
        }
    )
    has_traffic = False
    try:
        last_pos = None
        for i in range(10):
            obs, _ = env.reset(seed=5)
            assert env.engine.traffic_manager.random_traffic
            new_pos = [v.position for v in env.engine.traffic_manager.traffic_vehicles]
            if len(new_pos) > 0:
                has_traffic = True
            if last_pos is not None and len(new_pos) == len(last_pos):
                assert sum(
                    [norm(lastp[0] - newp[0], lastp[1] - newp[1]) >= 0.5 for lastp, newp in zip(last_pos, new_pos)]
                ), [(lastp, newp) for lastp, newp in zip(last_pos, new_pos)]
            last_pos = new_pos
        assert has_traffic
    finally:
        env.close()


def test_random_lane_width():
    env = MetaDriveEnv(
        {
            "num_scenarios": 5,
            "traffic_density": .2,
            "traffic_mode": "trigger",
            "start_seed": 12,
            "random_lane_width": True,
        }
    )
    try:
        o, _ = env.reset(seed=12)
        old_config_1 = env.vehicle.lane.width
        env.reset(seed=15)
        old_config_2 = env.vehicle.lane.width
        env.reset(seed=13)
        env.reset(seed=12)
        new_config = env.vehicle.lane.width
        assert old_config_1 == new_config
        env.reset(seed=15)
        new_config = env.vehicle.lane.width
        assert old_config_2 == new_config
        assert old_config_1 != old_config_2
    finally:
        env.close()


def test_random_lane_num():
    env = MetaDriveEnv(
        {
            "num_scenarios": 5,
            "traffic_density": .2,
            "traffic_mode": "trigger",
            "start_seed": 12,
            "random_lane_num": True,
        }
    )
    try:
        o, _ = env.reset(seed=12)
        old_config_1 = env.vehicle.navigation.get_current_lane_num()
        env.reset(seed=15)
        old_config_2 = env.vehicle.navigation.get_current_lane_num()
        env.reset(seed=13)
        env.reset(seed=12)
        new_config = env.vehicle.navigation.get_current_lane_num()
        assert old_config_1 == new_config
        env.reset(seed=15)
        new_config = env.vehicle.navigation.get_current_lane_num()
        assert old_config_2 == new_config
        env.close()
        env.reset(seed=12)
        assert old_config_1 == env.vehicle.navigation.get_current_lane_num()
        env.reset(seed=15)
        assert old_config_2 == env.vehicle.navigation.get_current_lane_num()
    finally:
        env.close()


def test_random_vehicle_parameter():
    env = MetaDriveEnv(
        {
            "num_scenarios": 5,
            "traffic_density": .2,
            "traffic_mode": "trigger",
            "start_seed": 12,
            "random_agent_model": True
        }
    )
    try:
        o, _ = env.reset(seed=12)
        old_config_1 = env.vehicle.get_config(True)
        env.reset(seed=15)
        old_config_2 = env.vehicle.get_config(True)
        env.reset(seed=13)
        env.reset(seed=12)
        new_config = env.vehicle.get_config(True)
        assert recursive_equal(old_config_1, new_config)
        env.reset(seed=15)
        new_config = env.vehicle.get_config(True)
        assert recursive_equal(old_config_2, new_config)
    finally:
        env.close()


if __name__ == '__main__':
    test_map_random_seeding()
    # test_seeding()
    # test_random_lane_width()



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import setup_logger
from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.component.traffic_participants.cyclist import Cyclist

if __name__ == "__main__":
    # setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "traffic_mode": "hybrid",
            "start_seed": 22,
            "accident_prob": 1.0,
            # "_disable_detector_mask":True,
            # "debug_physics_world": True,
            "debug": True,
            # "global_light": True,
            # "debug_static_world": True,
            "cull_scene": False,
            # "image_observation": True,
            # "controller": "joystick",
            "manual_control": True,
            "use_render": True,
            "decision_repeat": 5,
            "need_inverse_traffic": False,
            "rgb_clip": True,
            "map": "XSS",
            # "agent_policy": IDMPolicy,
            "random_traffic": False,
            "random_lane_width": True,
            # "random_agent_model": True,
            "driving_reward": 1.0,
            "show_interface": True,
            "force_destroy": False,
            # "camera_dist": -1,
            # "camera_pitch": 30,
            # "camera_height": 1,
            # "camera_smooth": False,
            # "camera_height": -1,
            # "window_size": (2400, 1600),
            "show_coordinates": True,
            "vehicle_config": {
                "enable_reverse": True,
                # "show_lidar": True
                # "image_source": "depth_camera",
                # "random_color": True
                # "show_lidar": True,
                # "spawn_lane_index":("1r1_0_", "1r1_1_", 0),
                # "destination":"2R1_3_",
                # "show_side_detector": True,
                # "show_lane_line_detector": True,
                # "side_detector": dict(num_lasers=2, distance=50),
                # "lane_line_detector": dict(num_lasers=2, distance=50),
                # # "show_line_to_dest": True,
                # "show_dest_mark": True
            },
        }
    )
    import time

    start = time.time()
    o, _ = env.reset()
    obj_1 = env.engine.spawn_object(Pedestrian, position=[30, 0], heading_theta=0, random_seed=1)
    obj_2 = env.engine.spawn_object(Pedestrian, position=[30, 6], heading_theta=0, random_seed=1)
    c_1 = env.engine.spawn_object(Cyclist, position=[30, 8], heading_theta=0, random_seed=1)
    obj_1.set_velocity([1, 0], 1, in_local_frame=True)
    # obj_1.show_coordinates()
    obj_2.set_velocity([1, 0], 2, in_local_frame=True)
    c_1.set_velocity([3, 0], 2, in_local_frame=True)
    # obj_2.show_coordinates()
    env.vehicle.set_velocity([10, 0], in_local_frame=False)
    for s in range(1, 10000):
        # print(c_1.heading_theta)
        o, r, tm, tc, info = env.step(env.action_space.sample())
        # obj_1.set_velocity([1, 0], 2, in_local_frame=True)
        # obj_2.set_velocity([1, 0], 0.8, in_local_frame=True)
        if s == 300:
            obj_1.set_velocity([1, 0], 0, in_local_frame=True)
            # obj_2.set_velocity([1, 0], 0, in_local_frame=True)
        elif s == 500:
            obj_1.set_velocity([1, 0], 1, in_local_frame=True)
        # else:
        #     obj_1.set_velocity([1, 0], 1, in_local_frame=True)

        if 100 < s < 300:
            obj_2.set_velocity([1, 0], 2, in_local_frame=True)
        elif 500 > s > 300:
            # print("here stop")
            obj_2.set_velocity([1, 0], 0, in_local_frame=True)
        elif s >= 500:
            obj_2.set_velocity([1, 0], 2, in_local_frame=True)

        # else:
        # if s % 100 == 0:
        #     env.close()
        #     env.reset()
        # info["fuel"] = env.vehicle.energy_consumption
        env.render(
            text={
                "heading_diff": env.vehicle.heading_diff(env.vehicle.lane),
                "lane_width": env.vehicle.lane.width,
                "lateral": env.vehicle.lane.local_coordinates(env.vehicle.position),
                "current_seed": env.current_seed,
                "step": s,
            }
        )
        # if d:
        #     env.reset()
        # # assert env.observation_space.contains(o)
        # if (s + 1) % 100 == 0:
        #     # print(
        #         "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
        #             s + 1,f
        #             time.time() - start, (s + 1) / (time.time() - start)
        #         )
        #     )
        # if d:
        # #     # env.close()
        # #     # print(len(env.engine._spawned_objects))
        # env.reset()
